"""A python library for building data pipelines with Tamr"""
from . import data_io
from . import filesystem
from . import realtime
from . import models
from . import project
from . import utils
from . import workflow
from . import notifications
from . import enrichment
from . import dataset
from . import sysadmin

import logging

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "data_io",
    "dataset",
    "models",
    "filesystem",
    "realtime",
    "project",
    "utils",
    "workflow",
    "notifications",
    "enrichment",
    "sysadmin",
]



"""Capabilities around running concurrent workflows"""
from . import Graph
from . import Planner
from . import PlanNode
from . import PlanNodeStatus
from . import PlanStatus

__all__ = ["Graph", "Planner", "PlanNode", "PlanNodeStatus", "PlanStatus"]



import json
import logging
import time
from collections import defaultdict
from typing import Dict, List, Optional, Union

import requests
from tamr_unify_client import Client
from tamr_unify_client.mastering.project import MasteringProject
from tamr_unify_client.operation import Operation

from tamr_toolbox.models.data_type import JsonDict
from tamr_toolbox.utils.operation import from_resource_id
from tamr_toolbox.utils.version import requires_tamr_version

LOGGER = logging.getLogger(__name__)


def update_realtime_match_data(
    *,
    project: MasteringProject,
    do_update_clusters: bool = True,
    do_use_manual_clustering: bool = False,
    **options,
) -> Operation:
    """
    Updates data for RealTime match queries if needed, based on latest published clusters.

    Args:
        project: project to be updated
        do_update_clusters: whether to update clusters, default True
        do_use_manual_clustering: whether to use externally managed clustering, default False
        options: Options passed to underlying :class:`~tamr_unify_client.operation.Operation`

    Returns:
        an operation object describing the update operation

    Raises:
        RuntimeError: if update API call fails
    """

    # Make sure we have the original name of the project to use with the match endpoint
    project_name = _get_internal_project_name(project)

    url = (
        f"projects/{project_name}:updateLLM?updateClusters={do_update_clusters}"
        f"&useManualClustering={do_use_manual_clustering}"
    )

    try:
        response = project.client.post(url).successful()
    except requests.exceptions.HTTPError as e:
        message = f"Match data update for {project_name} failed at submission time: {e}"
        LOGGER.error(message)
        raise RuntimeError(message)
    operation_id = response.content.decode("latin1")
    operation = from_resource_id(project.client, job_id=operation_id)

    return operation.apply_options(**options)


def poll_realtime_match_status(
    *, project: MasteringProject, match_client: Client, num_tries: int = 10, wait_sec: int = 1
) -> bool:
    """
    Check if match service is queryable. Try up to `num_tries` times at 1 sec (or user-specified)
    interval.

    Args:
        project: the mastering project whose status to check
        match_client: a Tamr client set to use the port of the Match API
        num_tries: max number of times to poll endpoint, default 10
        wait_sec: number of seconds to wait between tries, default 1

    Returns:
        bool indicating whether project is queryable
    """

    project_name = _get_internal_project_name(project)
    url = f"/api/v1/projects/{project_name}:isQueryable"
    counter = 0

    # Poll endpoint to until project is queryable or num_tries reached
    while counter < num_tries:

        response = match_client.get(url)
        queryable = response.content == b"true"

        if queryable:
            break

        counter += 1
        time.sleep(wait_sec)  # call api at wait_sec interval if project isn't yet queryable

    return queryable


def match_query(
    *,
    project: MasteringProject,
    match_client: Client,
    records: List[JsonDict],
    type: str,
    primary_key: Optional[str] = None,
    batch_size: Optional[int] = None,
    min_match_prob: Optional[float] = None,
    max_num_matches: Optional[int] = None,
) -> Dict[Union[int, str], List[JsonDict]]:
    """
    Find the best matching clusters or records for each supplied record. Returns a dictionary where
    each key correpsonds to an input record and the value is a list of the RealTime match results
    for that record. An empty result list indicates a null response from matching (or no responses
    above the `min_match_prob`, if that parameter was supplied).

    Args:
        project: the mastering project to query for matches
        match_client: a Tamr client set to use the port of the Match API
        records: list of records to match
        type: one of "records" or  "clusters" -- whether to pull record or cluster matches
        primary_key: a primary key for the data; if supplied, this must be a field in input records
        batch_size: split input into this batch size for match query calls (e.g. to prevent network
            timeouts), default None sends a single query with all records
        min_match_prob: if set, only matches with probability above minimum will be returned,
            default None
        max_num_matches: if set, at most max_num_matches will be returned for each input record in
            records, default None

    Returns:
        Dict keyed by integers (indices of inputs), or by `primary_key` if `primary_key` is
        supplied, with value a list containing matched data

    Raises:
        ValueError: if match `type` is not "records" or "clusters", or if `batch_size` is invalid
        RuntimeError: if query fails
    """

    result_dict = defaultdict(lambda: [])  # dict which defaults to empty list to hold results

    project_name = _get_internal_project_name(project)
    url = f"/api/v1/projects/{project_name}:match?type={type}"

    # Set up keys to read results
    if type == "records":
        record_key = "queryRecordId"
        prob_key = "matchProbability"
    elif type == "clusters":
        record_key = "entityId"
        prob_key = "avgMatchProb"
    else:
        raise ValueError(f"Unsupported match type {type}.")

    # Check batch size and set if not supplied
    if batch_size is None:
        batch_size = len(records)
        if batch_size == 0:
            LOGGER.warn("No input supplied to match_query -- returning empty result.")
            return result_dict
    elif batch_size <= 0:
        raise ValueError(f"Batch size must be non-negative: received {batch_size}")

    # Split into batches and convert to match query format
    for j in range(0, len(records), batch_size):
        json_recs = _prepare_json(records[j : j + batch_size], primary_key=primary_key, offset=j)

        try:
            response = match_client.post(url, json=json_recs).successful()
        except requests.exceptions.HTTPError as e:
            message = f"RealTime match query failed: {e}"
            LOGGER.error(message)
            raise RuntimeError(message)

        # Process responses
        if response.content == b"":  # handle null response
            continue

        # If data was found, decode, identify source record, and add match to corresponding
        # list of results in the result dictionary
        for resp_block in response.content.decode("utf-8").split("\n"):
            if resp_block:
                result = json.loads(resp_block)
                index = int(result[record_key]) if primary_key is None else result[record_key]
                if max_num_matches and len(result_dict[index]) >= max_num_matches:
                    continue

                if min_match_prob and result[prob_key] < min_match_prob:
                    continue

                result_dict[index].append(result)

    return result_dict


@requires_tamr_version(min_version="2022.009.0")
def transform_and_match_query(
    *,
    project: MasteringProject,
    match_client: Client,
    records: List[JsonDict],
    type: str,
    primary_key: Optional[str] = None,
    batch_size: Optional[int] = None,
    min_match_prob: Optional[float] = None,
    max_num_matches: Optional[int] = None,
    default_source_name: Optional[str] = None,
) -> Dict[Union[int, str], List[JsonDict]]:
    """
    Find the best matching clusters or records for each supplied record. Returns a dictionary where
    each key correpsonds to an input record and the value is a list of the RealTime match results
    for that record. An empty result list indicates a null response from matching (or no responses
    above the `min_match_prob`, if that parameter was supplied). Will run schema mapping and
    transformations prior to realtime match. If LLT, is not enabled will just run default LLM with
    no transformation or schema mapping

    Args:
        project: the mastering project to query for matches
        match_client: a Tamr client set to use the port of the Match API
        records: list of records to match
        type: one of "records" or  "clusters" -- whether to pull record or cluster matches
        primary_key: a primary key for the data; if supplied, this must be a field in input records
        batch_size: split input into this batch size for match query calls (e.g. to prevent network
            timeouts), default None sends a single query with all records
        min_match_prob: if set, only matches with probability above minimum will be returned,
            default None
        max_num_matches: if set, at most max_num_matches will be returned for each input record in
            records, default None
        default_source_name: the default source name used for schema mapping in LLT, default None

    Returns:
        Dict keyed by integers (indices of inputs), or by `primary_key` if `primary_key` is
        supplied, with value a list containing matched data

    Raises:
        ValueError: if match `type` is not "records" or "clusters", or if `batch_size` is invalid
        RuntimeError: if query fails
    """

    result_dict = defaultdict(lambda: [])  # dict which defaults to empty list to hold results

    project_name = _get_internal_project_name(project)

    # Set up keys to read results
    if type == "records":
        url = f"/api/v1/projects/{project_name}:matchRecords?transform=true"
        record_key = "queryRecordId"
        prob_key = "matchProbability"
    elif type == "clusters":
        url = f"/api/v1/projects/{project_name}:matchClusters?transform=true"
        record_key = "entityId"
        prob_key = "avgMatchProb"
    else:
        raise ValueError(f"Unsupported match type {type}.")

    if default_source_name:
        url = url + f"&defaultSourceName={default_source_name}"
    # Check batch size and set if not supplied
    if batch_size is None:
        batch_size = len(records)
        if batch_size == 0:
            LOGGER.warn("No input supplied to match_query -- returning empty result.")
            return result_dict
    elif batch_size <= 0:
        raise ValueError(f"Batch size must be non-negative: received {batch_size}")

    # Split into batches and convert to match query format
    for j in range(0, len(records), batch_size):
        json_recs = _prepare_json(records[j : j + batch_size], primary_key=primary_key, offset=j)

        try:
            print(url)
            response = match_client.post(url, json=json_recs).successful()
            print(response.content)
        except requests.exceptions.HTTPError as e:
            message = f"RealTime match query failed: {e}"
            LOGGER.error(message)
            raise RuntimeError(message)

        # Process responses
        if response.content == b"":  # handle null response
            continue

        # If data was found, decode, identify source record, and add match to corresponding
        # list of results in the result dictionary
        for resp_block in response.content.decode("utf-8").split("\n"):
            if resp_block:
                result = json.loads(resp_block)
                index = int(result[record_key]) if primary_key is None else result[record_key]
                if max_num_matches and len(result_dict[index]) >= max_num_matches:
                    continue

                if min_match_prob and result[prob_key] < min_match_prob:
                    continue

                result_dict[index].append(result)

    return result_dict


def _prepare_json(
    records: List[JsonDict], *, primary_key: Union[str, None], offset: int
) -> List[JsonDict]:
    """
    Put records into JSON format expected by RealTime match endpoint

    Args:
        records: list of records to match
        primary_key: a primary key for the data; if not None, must be a field in the input records
        offset: offset to apply to generated integer `recordId` -- this is necessary for batching

    Returns:
        List of formatted records

    Raises:
        ValueError: if primary_key is supplied but some supplied record(s) do not have the
            primary_key field
    """

    if primary_key is not None:
        try:
            json_records = [{"recordId": rec.pop(primary_key), "record": rec} for rec in records]
        except KeyError:
            raise ValueError(f"Not all input records had a primary key field {primary_key}.")
    else:  # use integers as recordId
        json_records = [
            {"recordId": str(offset + k), "record": rec} for k, rec in enumerate(records)
        ]

    return json_records


def _get_internal_project_name(project: MasteringProject) -> str:
    """Get project's internal name (as opposed to the `displayName`, which can be changed by users)

    Args:
        project: the mastering project of which to retrieve the internal name

    Returns:
        the project's internal name
    """
    # Get usage data for unified dataset of input project
    usage_resp = project.unified_dataset().usage()
    # Get project internal name from first output_from_project_step of unified dataset useage
    name = usage_resp.usage.output_from_project_steps[0].project_name
    return name



"""Tasks related to running jobs for Tamr Golden Records projects"""
from typing import List
import logging

from tamr_unify_client.operation import Operation
from tamr_unify_client.project.resource import Project

from tamr_toolbox.models.project_type import ProjectType
from tamr_toolbox.utils import operation
from tamr_toolbox.utils.version import requires_tamr_version

LOGGER = logging.getLogger(__name__)


@requires_tamr_version(min_version="2020.004.0")
def _run_custom(
    project: Project,
    *,
    run_profile_golden_records: bool = False,
    run_update_golden_records: bool = False,
    run_publish_golden_records: bool = False,
    process_asynchronously: bool = False,
) -> List[Operation]:
    """Executes specified steps of a golden records project.

    Args:
        project: The target golden records project
        run_profile_golden_records: Whether reprofile should be called on the golden records
        project
        run_update_golden_records: Whether refresh should be called on the draft golden records
            dataset
        run_publish_golden_records: Whether refresh should be called on the published golden
            records dataset
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run

    Raises:
        TypeError: if the `project` is not a Golden Record project
    """
    if ProjectType[project.type] != ProjectType.GOLDEN_RECORDS:
        error_msg = f"Cannot use as a golden records project. Project type: {project.type}"
        LOGGER.error(error_msg)
        raise TypeError(error_msg)

    completed_operations = []
    if run_profile_golden_records:
        LOGGER.info(
            f"Updating all profiling information for golden records project {project.name} "
            f"(id={project.resource_id})."
        )
        resp = project.client.post(
            f"/api/versioned/v1/projects/{project.resource_id}/goldenRecordsProfile:refresh"
        ).successful()

        op = Operation.from_response(client=project.client, response=resp)
        if not process_asynchronously:
            op = op.wait()
            operation.enforce_success(op)
        completed_operations.append(op)

    if run_update_golden_records:
        LOGGER.info(
            f"Updating the draft golden records for project {project.name} "
            f"(id={project.resource_id})."
        )
        resp = project.client.post(
            f"/api/versioned/v1/projects/{project.resource_id}/goldenRecords:refresh"
        ).successful()

        op = Operation.from_response(client=project.client, response=resp)
        if not process_asynchronously:
            op = op.wait()
            operation.enforce_success(op)
        completed_operations.append(op)

    if run_publish_golden_records:
        LOGGER.info(
            f"Publishing golden records for project {project.name} (id={project.resource_id})."
        )
        resp = project.client.post(
            f"/api/versioned/v1/projects/{project.resource_id}/publishedGoldenRecords:refresh"
            f"?validate=true&version=CURRENT"
        ).successful()

        op = Operation.from_response(client=project.client, response=resp)
        if not process_asynchronously:
            op = op.wait()
            operation.enforce_success(op)
        completed_operations.append(op)

    return completed_operations


def run(project: Project, *, process_asynchronously: bool = False) -> List[Operation]:
    """Run the project

    Args:
        project: Target golden records project
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project,
        run_profile_golden_records=True,
        run_update_golden_records=True,
        run_publish_golden_records=True,
        process_asynchronously=process_asynchronously,
    )


def update_profiling_info(
    project: Project, *, process_asynchronously: bool = False
) -> List[Operation]:
    """Updating all profiling information for golden records project

    Args:
        project: Target golden records project
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project,
        run_profile_golden_records=True,
        run_update_golden_records=False,
        run_publish_golden_records=False,
        process_asynchronously=process_asynchronously,
    )


def update_golden_records(
    project: Project, *, process_asynchronously: bool = False
) -> List[Operation]:
    """Update the draft golden records of a project

    Args:
        project: Target golden records project
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project,
        run_profile_golden_records=False,
        run_update_golden_records=True,
        run_publish_golden_records=False,
        process_asynchronously=process_asynchronously,
    )


def publish_golden_records(
    project: Project, *, process_asynchronously: bool = False
) -> List[Operation]:
    """Publish the  golden records of a project

    Args:
        project: Target golden records project
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project,
        run_profile_golden_records=False,
        run_update_golden_records=False,
        run_publish_golden_records=True,
        process_asynchronously=process_asynchronously,
    )


def update_and_publish(
    project: Project, *, process_asynchronously: bool = False
) -> List[Operation]:
    """Update the draft golden records and publish the golden records

    Args:
        project: Target golden records project
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project,
        run_profile_golden_records=False,
        run_update_golden_records=True,
        run_publish_golden_records=True,
        process_asynchronously=process_asynchronously,
    )



"""Tasks related to enriching data out of Tamr"""

from . import address_mapping
from . import address_validation
from . import dictionary
from . import enrichment_utils
from . import translate
from . import api_client

__all__ = [
    "address_mapping",
    "address_validation",
    "dictionary",
    "enrichment_utils",
    "translate",
    "api_client",
]



"""Tasks related to scripting"""
from . import client
from . import config
from . import downstream
from . import logger
from . import operation
from . import testing
from . import upstream
from . import version
from . import custom_button


__all__ = [
    "client",
    "config",
    "downstream",
    "logger",
    "operation",
    "testing",
    "upstream",
    "version",
    "custom_button",
]



"""Tasks related to efficiently translating data not present in existing translation
dictionaries"""
import logging
import math
import os
from typing import Dict, List, Union

from tamr_toolbox.enrichment.api_client import google_translate
from tamr_toolbox.enrichment.dictionary import TranslationDictionary, save, update
from tamr_toolbox.enrichment.enrichment_utils import _yield_chunk

# Building our documentation requires access to all dependencies, including optional ones
# This environments variable is set automatically when `invoke docs` is used
BUILDING_DOCS = os.environ.get("TAMR_TOOLBOX_DOCS") == "1"
if BUILDING_DOCS:
    # Import relevant optional dependencies
    from google.cloud.translate_v2 import Client as GoogleTranslateClient


LOGGER = logging.getLogger(__name__)


def _filter_numeric_and_null_phrases(phrase: Union[str, None]) -> str:
    """
    Transform None and numbers saved as text as empty strings

    Args:
        phrase: data to filter

    Returns:
        An empty string
        Raises an error if the input is neither None nor a string

    Raises:
        TypeError: is the provided phrase is not of type string
    """
    if phrase is None:
        return ""
    elif not isinstance(phrase, str):
        error_message = (
            f"{phrase} is not in text format. " f"Only text can be translated, check data type."
        )
        LOGGER.error(error_message)
        raise TypeError(error_message)
    elif phrase.isnumeric():
        return ""
    else:
        return phrase


def standardize_phrases(original_phrases: List[str]) -> List[str]:
    """
    Standardize phrases to translate to avoid re-translating previously translated phrases but
    with different formating

    Args:
        original_phrases: List of phrases to standardize

    Returns:
        List of standardized text
    """
    standardized_phrases = [
        " ".join(_filter_numeric_and_null_phrases(phrase).lower().split())
        for phrase in original_phrases
    ]
    return standardized_phrases


def get_phrases_to_translate(
    original_phrases: List[str], translation_dictionary: Dict[str, TranslationDictionary]
) -> List[str]:
    """
    Find phrases not previously translated and initiate dictionary entry

    Args:
        original_phrases: list of phrases to translate
        translation_dictionary: a translation dictionary

    Returns:
        List of standardized phrases not present as keys of the translation dictionary

    """
    count_already_translated = 0
    count_needing_translation = 0

    for original, standard in zip(original_phrases, standardize_phrases(original_phrases)):
        if standard in translation_dictionary.keys():
            translation_dictionary[standard].original_phrases.add(original)
            count_already_translated += 1
        else:
            translation_dictionary[standard] = TranslationDictionary(
                original_phrases={original}, standardized_phrase=standard
            )
            count_needing_translation += 1

    LOGGER.info(
        f"From the {len(original_phrases)} sent for translation, "
        f"{count_already_translated} can be translated with the dictionary and "
        f"{count_needing_translation} need to be translated"
    )

    to_translate = [
        t.standardized_phrase
        for t in translation_dictionary.values()
        if t.translated_phrase is None
    ]
    LOGGER.debug(f"{to_translate}")
    return to_translate


def from_list(
    all_phrases: List[str],
    client: "GoogleTranslateClient",
    dictionary: Dict[str, TranslationDictionary],
    *,
    source_language: str = "auto",
    target_language: str = "en",
    chunk_size: int = 100,
    translation_model: str = "nmt",
    intermediate_save_every_n_chunks: Union[int, None] = None,
    intermediate_save_to_disk: bool = False,
    intermediate_folder: str = "/tmp",
) -> Dict[str, TranslationDictionary]:
    """
    Translate a list of phrases from source language to target language.
    The translation is saved in a dictionary on your local file system before updating the
    main dictionary

    Args:
        all_phrases: List of standardized phrases to translate.
        client: a google translate api client
        dictionary: a toolbox translation dictionary
        source_language: the language the text to translate is in, "auto" means the api_client
            google_api api_client will try to detect the source language automatically
        target_language: the language to translate into
        chunk_size: number of phrases to translate per api_client calls, set too high and you
            will hit API user rate limit errors
        translation_model: google_api api_client api_client model to use, "nmt" or "pbmt".
            Choose "pbmt" if an "nmt" model doesn't exists for your source to target language pair
        intermediate_save_every_n_chunks: save periodically api_client dictionary to disk every n
            chunk of phrases translated
        intermediate_save_to_disk: decide whether to save periodically the dictionary to disk to
            avoid loss of translation data if code breaks
        intermediate_folder: path to folder where dictionary will be save periodically to avoid
            loss of translation data

    Returns:
        The updated translation dictionary

    Raises:
        ValueError: if the argument chunk_size is set to 0
    """
    if chunk_size == 0:
        error_message = "Translation chunk size cannot be of size 0"
        LOGGER.error(error_message)
        raise ValueError(error_message)

    if intermediate_save_every_n_chunks == 0 or intermediate_save_every_n_chunks is None:
        intermediate_save_every_n_chunks = math.inf

    unique_all_phrases = list(set(all_phrases))
    nbr_of_unique_phrases = len(unique_all_phrases)

    phrases_to_translate = get_phrases_to_translate(unique_all_phrases, dictionary)
    number_phrases_to_translate = len(phrases_to_translate)

    if number_phrases_to_translate == 0:
        LOGGER.info("All phrases to translate are found in the local dictionary.")

    else:
        LOGGER.info(
            f"Of the {nbr_of_unique_phrases} unique phrases to translate, "
            f"{number_phrases_to_translate} were not found in the dictionary."
        )

        # Google has a translation rate limits
        # to avoid hitting those the phrases are sent for translation in chunks
        number_of_chunks = math.ceil(number_phrases_to_translate / chunk_size)

        tmp_dictionary = {}
        for ix, chunk_of_phrases in enumerate(_yield_chunk(phrases_to_translate, chunk_size)):
            LOGGER.debug(f"Translating chunk {ix + 1} out of {number_of_chunks}.")
            translated_phrases = google_translate.translate(
                phrases_to_translate=chunk_of_phrases,
                client=client,
                source_language=source_language,
                target_language=target_language,
                translation_model=translation_model,
            )
            if translated_phrases is not None:
                tmp_dictionary.update(translated_phrases)

            if (ix % intermediate_save_every_n_chunks) == 0:
                LOGGER.info("Saving intermediate outputs")
                update(main_dictionary=dictionary, tmp_dictionary=tmp_dictionary)
                if intermediate_save_to_disk:
                    save(
                        translation_dictionary=dictionary,
                        dictionary_folder=intermediate_folder,
                        target_language=target_language,
                        source_language=source_language,
                    )
                # resetting temporary results after saving it
                tmp_dictionary = {}

        # update dictionary
        update(main_dictionary=dictionary, tmp_dictionary=tmp_dictionary)

    return dictionary



"""Tasks related to project movement as part of Tamr projects"""
from typing import List, Optional
import logging
from urllib import request

from tamr_unify_client.project.resource import Project
from tamr_unify_client import Client

from tamr_toolbox import utils
from tamr_toolbox.utils.operation import Operation
from tamr_toolbox.utils.version import requires_tamr_version

LOGGER = logging.getLogger(__name__)


@requires_tamr_version(min_version="2021.005.0")
def export_artifacts(
    *,
    project: Project,
    artifact_directory_path: str,
    exclude_artifacts: Optional[List[str]] = None,
    asynchronous: bool = False,
) -> Operation:
    """Export project artifacts for project movement

    Requires Tamr 2021.005.0 or later

    Args:
        project: a tamr project object
        artifact_directory_path: export directory for project artifacts
        exclude_artifacts: list of artifacts to exclude
        asynchronous: flag to run function asynchronously
    Returns:
        operation for project export api call
    """
    # initializing empty lists if any
    if exclude_artifacts is None:
        exclude_artifacts = []

    # make project export api request
    body = {"artifactDirectory": artifact_directory_path, "excludeArtifacts": exclude_artifacts}
    LOGGER.info(f"Starting to export {project.name} of type {project.type}.")
    response = project.client.post(
        f"/api/versioned/v1/projects/{project.resource_id}:export", json=body
    )
    # Raise error if export api call not successful
    if not response.ok:
        error_message = f"Error exporting project artifacts: {response.json()}"
        LOGGER.error(error_message)
        raise ValueError(error_message)

    # get operation
    job_id = response.json()["id"]
    operation = utils.operation.from_resource_id(tamr=project.client, job_id=job_id)
    if not asynchronous:
        # periodically check export job until completion
        LOGGER.info(f"Waiting for export to be created.")
        operation = operation.wait()
        utils.operation.enforce_success(operation)
        LOGGER.info(f"Export complete: {operation}.")
    else:
        LOGGER.info(f"Export called asynchronously: {operation}.")

    return operation


@requires_tamr_version(min_version="2021.005.0")
def import_artifacts(
    *,
    project_artifact_path: str,
    tamr_client: Client,
    target_project: Project = None,
    new_project_name: str = None,
    new_unified_dataset_name: Optional[str] = None,
    exclude_artifacts: Optional[List[str]] = None,
    include_additive_artifacts: Optional[List[str]] = None,
    include_destructive_artifacts: Optional[List[str]] = None,
    fail_if_not_present: bool = False,
    asynchronous: bool = False,
    overwrite_existing: bool = False,
) -> Operation:
    """Import project artifacts into a tamr instance

    Requires Tamr 2021.005.0 or later

    Args:
        tamr_client: a tamr client
        project_artifact_path: project artifacts zip filepath
        target_project: an optional target project for migration
        new_project_name: new project name
        new_unified_dataset_name: new unified dataset name
        exclude_artifacts: list of artifacts to exclude in import
        include_additive_artifacts: list of artifacts to import only additively
        include_destructive_artifacts: list of artifacts to import destructively
        fail_if_not_present: flag to fail project if not already present in instance
        asynchronous: flag to run function asynchronously
        overwrite_existing: flag to overwrite existing project artifacts
    Returns:
        operation for project import api call
    """
    # initializing empty lists if any
    if exclude_artifacts is None:
        exclude_artifacts = []
    if include_additive_artifacts is None:
        include_additive_artifacts = []
    if include_destructive_artifacts is None:
        include_destructive_artifacts = []

    # handle spaces in project artifact path
    if " " in project_artifact_path:
        project_artifact_path = request.pathname2url(project_artifact_path)

    # make project import api request
    body = {
        "projectArtifact": project_artifact_path,
        "excludeArtifacts": exclude_artifacts,
        "includeAdditiveArtifacts": include_additive_artifacts,
        "includeDestructiveArtifacts": include_destructive_artifacts,
        "failIfNotPresent": fail_if_not_present,
    }

    if target_project:
        if new_project_name or new_unified_dataset_name:
            error_message = (
                "Cannot import to existing project and simultaneously set "
                "new_project_name or new_unified_dataset_name."
            )
            LOGGER.error(error_message)
            raise KeyError(error_message)
        if overwrite_existing:
            LOGGER.info(
                f"Starting to import artifacts from {project_artifact_path} "
                f"into existing project {target_project.name}"
                f"with id {target_project.resource_id}."
            )
            response = tamr_client.post(
                f"/api/versioned/v1/projects/{target_project.resource_id}:import", json=body
            )
        else:
            error_message = "Unable to overwrite existing project; overwrite flag is off."
            LOGGER.error(error_message)
            raise KeyError(error_message)
    else:
        if new_project_name in [p.name for p in tamr_client.projects]:
            error_message = f"New project name {new_project_name} already exists."
            LOGGER.error(error_message)
            raise ValueError(error_message)
        if new_unified_dataset_name in [ds.name for ds in tamr_client.datasets]:
            # otherwise response will be ok, but operation 'state'='FAILED'
            # and artifacts will be partially migrated : creates issues on re-run.
            error_message = f"New unified dataset name {new_unified_dataset_name} already exists."
            LOGGER.error(error_message)
            raise ValueError(error_message)
        body["newProjectName"] = new_project_name
        body["newUnifiedDatasetName"] = new_unified_dataset_name
        LOGGER.info(
            f"Starting to import artifacts from {project_artifact_path} "
            f"into new project {new_project_name}."
        )
        response = tamr_client.post("/api/versioned/v1/projects:import", json=body)

    # Raise exception if import was not successful
    if not response.ok:
        error_message = f"Error importing project artifacts: {response.json()}"
        LOGGER.error(error_message)
        raise ValueError(error_message)

    # get operation
    job_id = response.json()["id"]
    operation = utils.operation.from_resource_id(tamr=tamr_client, job_id=job_id)
    if not asynchronous:
        # periodically check export job until completion
        LOGGER.info(f"Waiting for project to be imported.")
        operation = operation.wait()
        utils.operation.enforce_success(operation)
        LOGGER.info(f"Import complete: {operation}.")
    else:
        LOGGER.info(f"Import called asynchronously: {operation}.")

    return operation



"""Tasks related to schema mapping as part of Tamr projects"""
from tamr_unify_client.project.attribute_mapping.resource import (
    AttributeMappingSpec,
    AttributeMapping,
)
from tamr_unify_client.project.resource import Project
from tamr_unify_client.dataset.resource import Dataset

from typing import List
import logging
from json import JSONDecodeError

LOGGER = logging.getLogger(__name__)


def _get_mapping_spec_for_ud(
    *, source_ds_name: str, source_attr_name: str, unified_attr_name: str, unified_ds_name: str
) -> AttributeMappingSpec:
    """Makes an AttributeMappingSpec to be used to map attributes.
    Is a wrapper for its creation to hide internals in one place.

    Args:
        source_ds_name: the name of the source dataset
        source_attr_name: the name of the source attribute
        unified_attr_name: the name of the unified attribute
        unified_ds_name: the name of the unified dataset

    Returns:
        AttributeMappingSpec containing the passed information
    """
    spec = (
        AttributeMappingSpec.new()
        .with_input_attribute_name(source_attr_name)
        .with_input_dataset_name(source_ds_name)
        .with_unified_dataset_name(unified_ds_name)
        .with_unified_attribute_name(
            unified_attr_name
        )  # rest can be empty - Tamr will fill them in
        .with_input_attribute_id("")
        .with_relative_input_attribute_id("")
        .with_unified_attribute_id("")
        .with_relative_unified_attribute_id("")
    )
    return spec


def map_attribute(
    project: Project,
    *,
    source_attribute_name: str,
    source_dataset_name: str,
    unified_attribute_name: str,
) -> AttributeMapping:
    """
    Maps source_attribute in source_dataset to unified_attribute in unified_dataset.
    If the mapping already exists it will log
    a warning and return the existing AttributeMapping from the project's collection.

    Args:
        source_attribute_name: Source attribute name to map
        source_dataset_name: Source dataset containing the source attribute
        unified_attribute_name: Unified attribute to which to map the source attribute
        project: The project in which to perform the mapping

    Returns:
        The created AttributeMapping

    Raises:
        ValueError: if input variables `source_attribute_name` or `source_dataset_name` or
            `unified_attribute_name` are set to empty strings;
            or if the dataset `source_dataset_name` is not found on Tamr;
            or if `source_attribute_name` is missing from the attributes of `source_attribute_name`
    """
    # simple validation, nothing should be empty
    variables = [source_attribute_name, source_dataset_name, unified_attribute_name]
    empty_variables = [x for x in variables if x == ""]
    if empty_variables:
        empty_variable_string = ", ".join(empty_variables)
        error_message = (
            f"The following variables are set to empty strings and "
            f"need to be filled in: {empty_variable_string} !"
        )
        LOGGER.error(error_message)
        raise ValueError(error_message)

    # also validate that the dataset exists and has this column
    try:
        source_dataset = project.client.datasets.by_name(source_dataset_name)
    except KeyError:
        error_msg = f"Dataset {source_dataset_name} not found!"
        LOGGER.error(error_msg)
        raise ValueError(error_msg)

    if source_attribute_name not in [x.name for x in source_dataset.attributes]:
        error_msg = f"Attribute {source_attribute_name} not found in {source_dataset_name}!"
        LOGGER.error(error_msg)
        raise ValueError(error_msg)

    # generate mapping spec
    mapping_spec = _get_mapping_spec_for_ud(
        source_attr_name=source_attribute_name,
        source_ds_name=source_dataset_name,
        unified_attr_name=unified_attribute_name,
        unified_ds_name=project.unified_dataset().name,
    )

    # add the mapping to the project's collection - this is what does the actual mapping
    try:
        return project.attribute_mappings().create(mapping_spec.to_dict())
    except JSONDecodeError as e:
        # can get a jsondecode error if the attribute is already mapped.
        # If it is, then log a warning and return empty mapping
        # if it is not already mapped break loudly
        m: AttributeMapping
        for m in project.attribute_mappings().stream():
            if (
                m.input_dataset_name == source_dataset_name
                and m.input_attribute_name == source_attribute_name
                and m.unified_attribute_name == unified_attribute_name
            ):
                # mapping exists, log warning and return existing mapping
                LOGGER.warning(
                    f"mapping of attribute {source_attribute_name} in dataset "
                    f"{source_dataset_name} to unified attribute {unified_attribute_name} "
                    f"already exists! Returning existing mapping spec"
                )
                return m

        # if haven't returned then throw the JSONDecodeError
        raise e


def unmap_attribute(
    project: Project,
    *,
    source_attribute_name: str,
    source_dataset_name: str,
    unified_attribute_name: str,
) -> None:
    """
    Unmaps a source attribute.

    Args:
        source_attribute_name: the name of the source attribute to unmap
        source_dataset_name: the name of the source dataset containing that source attribute
        unified_attribute_name: the unified attribute from which to unmap
        project: the project in which to unmap the attribute

    Returns:
        None
    """

    LOGGER.info(
        f"Trying to remove mapping of source attribute {source_attribute_name} in dataset "
        f"{source_dataset_name} from unified attribute {unified_attribute_name}"
    )

    # get mapping collection
    mapping_collection = project.attribute_mappings()

    # run through and get the resource id of the mapping to remove
    resource_id_to_remove = None
    for mapping in mapping_collection.stream():
        # consider it match if all of source attribute, source dataset and unified attribute
        # are equal
        if (
            source_attribute_name == mapping.input_attribute_name
            and source_dataset_name == mapping.input_dataset_name
            and unified_attribute_name == mapping.unified_attribute_name
        ):
            resource_id_to_remove = mapping.resource_id
            break

    # log warning if resource id wasn't found
    if resource_id_to_remove is None:
        LOGGER.warning(
            f"Mapping of {source_attribute_name} in dataset {source_dataset_name} to "
            f"unified attribute {unified_attribute_name} not found!"
        )
    # if found remove it
    else:
        mapping_collection.delete_by_resource_id(resource_id_to_remove)


def bootstrap_dataset(
    project: Project, *, source_dataset: Dataset, force_add_dataset_to_project: bool = False
) -> List[AttributeMapping]:
    """
    Bootstraps a dataset (i.e. maps all source columns to themselves)

    Args:
        source_dataset: the source dataset (a Dataset object not a string)
        project: the project to do the mapping ing
        force_add_dataset_to_project: boolean whether to add the dataset to the project
            if it is not already a part of it

    Returns:
        List of the AttributeMappings generated

    Raises:
        RuntimeError: if `source_dataset` is not part of the given `project`, set
            'force_add_dataset_to_project' flag to True to automatically add it
    """

    # check if dataset is in the project - python doesn't handle comparison of Dataset objects
    # well so check on name
    if source_dataset.name not in [x.name for x in project.input_datasets()]:
        if force_add_dataset_to_project:
            LOGGER.info(f"adding dataset {source_dataset.name} to project {project.name}")
            project.add_input_dataset(source_dataset)
        else:
            raise RuntimeError(
                f"dataset {source_dataset.name} not in project {project.name}!"
                + "Set 'force_add_dataset_to_project' flag to True to automatically add it"
            )

    # for each attribute map it
    source_dataset_name = source_dataset.name
    completed_mappings = []
    for attribute in source_dataset.attributes:
        attribute_name = attribute.name
        mapping = map_attribute(
            source_attribute_name=attribute_name,
            source_dataset_name=source_dataset_name,
            unified_attribute_name=attribute_name,
            project=project,
        )
        completed_mappings.append(mapping)

    return completed_mappings


def unmap_dataset(
    project: Project,
    *,
    source_dataset: Dataset,
    remove_dataset_from_project: bool = False,
    skip_if_missing: bool = False,
) -> None:
    """
    Wholly unmaps a dataset and optionally removes it from a project.

    Args:
        source_dataset: the source dataset (Dataset object not a string) to unmap
        project: the project in which to unmap the dataset
        remove_dataset_from_project: boolean to also remove the dataset from the project
        skip_if_missing: boolean to skip if dataset is not in project. If set to false and
            dataset is not in project will raise a RuntimeError

    Returns:
        None

    Raises:
        RuntimeError: if `source_dataset` is not in `project` and `skip_if_missing` not set to True
    """

    # check to make sure dataset is in project and log a warning if it is not
    if source_dataset.name not in [x.name for x in project.input_datasets()]:
        if skip_if_missing:
            LOGGER.warning(
                f"Dataset to unmap {source_dataset.name} not in project {project.name}! "
                f"However skip_if_missing flag is set so will do nothing"
            )
            return None
        else:
            error_message = (
                f"Dataset to unmap {source_dataset.name} not in project "
                f"{project.name} and skip_if_missing not set to True so failing! "
            )
            LOGGER.error(error_message)
            raise RuntimeError(error_message)

    # the resource ids of attribute mappings unfortunately change when you delete one
    # so need to just do this until there are no mappings left for the source dataset of interest
    while True:
        mappings = [
            x
            for x in project.attribute_mappings().stream()
            if x.input_dataset_name == source_dataset.name
        ]
        # if no mappings found for this dataset then break
        if not mappings:
            break
        for mapping in mappings:
            # can only delete one then have to break out of inner loop
            project.attribute_mappings().delete_by_resource_id(mapping.resource_id)
            break

    # optionally remove dataset from the project
    if remove_dataset_from_project:
        project.remove_input_dataset(source_dataset)



"""Enum representing project types in Tamr"""
from enum import Enum


class ProjectType(Enum):
    """A dataclass representing the project type codes in Tamr

    Args:
        SCHEMA_MAPPING_RECOMMENDATIONS: Schema Mapping Project Type
        DEDUP: Mastering Project Type
        GOLDEN_RECORDS: Golden Record Project Type
        CATEGORIZATION: Categorization Project Type
    """

    SCHEMA_MAPPING_RECOMMENDATIONS = "SCHEMA_MAPPING_RECOMMENDATIONS"
    DEDUP = "DEDUP"
    GOLDEN_RECORDS = "GOLDEN_RECORDS"
    CATEGORIZATION = "CATEGORIZATION"
    ENRICHMENT = "ENRICHMENT"



"""Tasks related to a Tamr instance"""
import re
from typing import Optional, Dict, List, Any

import subprocess
import logging
import time
import os

import yaml

LOGGER = logging.getLogger(__name__)

# Building our documentation requires access to all dependencies, including optional ones
# This environments variable is set automatically when `invoke docs` is used
BUILDING_DOCS = os.environ.get("TAMR_TOOLBOX_DOCS") == "1"
if BUILDING_DOCS:
    # Import relevant optional dependencies
    import paramiko


def _run_remote_command(
    command: str,
    *,
    remote_client: "paramiko.SSHClient",
    command_input: Optional[bytes] = None,
    verbose=False,
) -> (int, str, str):
    """Runs the provided command in a remote environment using the provided ssh client

    Args:
        command: The command to run
        remote_client: An ssh client providing a remote connection
        command_input: Content to send to stdin after command is started
        verbose: Whether the full command, stdout, and stderr should be logged at INFO level

    Returns:
        (exit code of command, stdout of command, stderr of command)

    """

    command_details_message = (
        f"Running command [{command}] on "
        f"remote machine ({remote_client.get_transport().getpeername()[0]}) "
        f"as user '{remote_client.get_transport().get_username()}'."
    )
    if verbose:
        LOGGER.info(command_details_message)
    else:
        LOGGER.debug(command_details_message)

    # Initiate command
    stdin_file, stdout_file, stderr_file = remote_client.exec_command(command)
    command_channel = stdout_file.channel

    # Provide input to command (such as a password) if given
    if command_input is not None:
        stdin_file.write(command_input)
    stdin_file.close()

    # While waiting for command to complete,
    # Collect command output
    # Some bash scripts do not close their stderr and stdout even after the exit status is ready
    # This leads to calls like stdout_file.readlines() to hang indefinitely
    # We empty the buffer instead to avoid this issue
    full_stdout = ""
    full_stderr = ""
    while True:
        time.sleep(1)
        stderr = command_channel.in_stderr_buffer.empty().decode("utf-8")
        stdout = command_channel.in_buffer.empty().decode("utf-8")
        if len(stdout) > 0:
            full_stdout += stdout
            if verbose:
                LOGGER.info(f"STDOUT: {stdout}")
            else:
                LOGGER.debug(f"STDOUT: {stdout}")
        if len(stderr) > 0:
            full_stderr += stderr
            if verbose:
                LOGGER.info(f"STDERR: {stderr}")
            else:
                LOGGER.debug(f"STDERR: {stderr}")

        if command_channel.exit_status_ready():
            stdout_file.close()
            stderr_file.close()
            break

    command_output_message = f"Command ended with exit code {command_channel.exit_status}."
    if verbose:
        LOGGER.info(command_output_message)
    else:
        LOGGER.debug(command_output_message)

    return command_channel.exit_status, full_stdout, full_stderr


def _run_local_command(
    command: str, *, command_input: Optional[bytes] = None, verbose=False
) -> (int, str, str):
    """Runs the provided command in the local shell

    Args:
        command: The command to run
        command_input: Content to send to stdin after command is started
        verbose: Whether the full command, stdout, and stderr should be logged at INFO level
    Returns:
        (exit code of command, stdout of command, stderr of command)

    """
    command_details_message = f"Running command [{command}] on local machine."
    if verbose:
        LOGGER.info(command_details_message)
    else:
        LOGGER.debug(command_details_message)

    # Initiate command
    process = subprocess.Popen(
        command, stderr=subprocess.PIPE, stdout=subprocess.PIPE, stdin=subprocess.PIPE, shell=True
    )

    # Provide input to command (such as a password) if given
    if command_input is not None:
        process.stdin.write(command_input)
    process.stdin.close()

    # Wait for command to complete
    # Collect command output
    # Some bash scripts do not close their stderr and stdout even after the exit status is ready
    # This leads to calls like stdout.readlines() to hang indefinitely
    # We use peek instead to avoid this issue
    full_stdout = ""
    full_stderr = ""
    while True:
        time.sleep(1)
        stdout = process.stdout.peek().decode("utf-8")
        stderr = process.stderr.peek().decode("utf-8")
        if len(stdout) > 0:
            full_stdout += stdout
            if verbose:
                LOGGER.info(f"STDOUT: {stdout}")
            else:
                LOGGER.debug(f"STDOUT: {stdout}")
        if len(stderr) > 0:
            full_stderr += stderr
            if verbose:
                LOGGER.info(f"STDERR: {stderr}")
            else:
                LOGGER.debug(f"STDERR: {stderr}")

        if process.poll() is not None:
            break

    command_output_message = f"Command ended with exit code {process.returncode}."
    if verbose:
        LOGGER.info(command_output_message)
    else:
        LOGGER.debug(command_output_message)

    return process.returncode, full_stdout, full_stderr


def _run_command(
    command: str,
    *,
    remote_client: Optional["paramiko.SSHClient"] = None,
    impersonation_username: Optional[str] = None,
    impersonation_password: Optional[str] = None,
    enforce_success: bool = True,
    verbose=False,
) -> (int, str, str):
    """Runs the provided command in a remote environment if an ssh client is specified otherwise
    run the provided command in the local shell

    If an impersonation_username is provided, the command is run as the provided user.
    If an impersonation_password is provided, password authentication is used for impersonation,
    otherwise sudo is used

    Args:
        command: The command to run
        remote_client: An ssh client providing a remote connection
        impersonation_username: A bash user to run the command as
        impersonation_password: The password for the impersonation_username
        enforce_success: Whether to throw an error if the command fails
        verbose: Whether the full command, stdout, and stderr should be logged at INFO level

    Returns:
        (exit code of command, stdout of command, stderr of command)

    Raises:
        RuntimeError: Raised when enforce_success is True and the command exists with an non-0 code

    """
    # If a username is specified, switch users for running the command
    if impersonation_username is not None:
        if impersonation_password is None:
            # If no password is provided use sudo
            command = f"sudo su - {impersonation_username} -c '{command}'"
            password_input = None
        else:
            # If a password is required pass it into the running process
            command = f"su - {impersonation_username} -c '{command}'"
            password_input = impersonation_password.encode("utf-8")
    else:
        # If no username is specified, no password is needed
        password_input = None

    # Run the command on the local or remote system
    if remote_client is None:
        exit_code, stdout, stderr = _run_local_command(
            command, command_input=password_input, verbose=verbose
        )
    else:
        exit_code, stdout, stderr = _run_remote_command(
            command, remote_client=remote_client, command_input=password_input, verbose=verbose
        )

    # When enforce_success is True, raise an error for non-zero exit codes
    if enforce_success and exit_code != 0:
        if remote_client is None:
            task_description = f"local command."
        else:
            remote_ip = remote_client.get_transport().getpeername()[0]
            remote_username = remote_client.get_transport().get_username()
            task_description = f"Failed to run remote command on {remote_ip} as {remote_username}."
        raise RuntimeError(
            f"{task_description} "
            f"Command: [{command}] exited with code {exit_code}.\n"
            f"STDOUT: [{stdout.strip()}]\n"
            f"STDERR: [{stderr.strip()}]"
        )
    return exit_code, stdout, stderr


def start_tamr(
    *,
    tamr_install_dir: str,
    include_dependencies: bool = True,
    remote_client: Optional["paramiko.SSHClient"] = None,
    impersonation_username: Optional[str] = None,
    impersonation_password: Optional[str] = None,
    verbose=False,
) -> None:
    """Starts the Tamr software and the Tamr dependencies if include_dependencies is true.

    Runs in a remote environment if an ssh client is specified otherwise runs in the local shell.
    If an impersonation_username is provided, the command is run as the provided user.
    If an impersonation_password is provided, password authentication is used for impersonation,
    otherwise sudo is used.

    Args:
        tamr_install_dir: Full path to directory where Tamr is installed
        include_dependencies: Whether Tamr dependencies should be started
        remote_client: An ssh client providing a remote connection
        impersonation_username: A bash user to run the command as,
            this should be the tamr install user
        impersonation_password: The password for the impersonation_username
        verbose: Whether the full command, stdout, and stderr should be logged at INFO level

    Returns:
        None

    Raises:
        RuntimeError: Raised if Tamr start script(s) fail

    """
    if include_dependencies:
        LOGGER.info(f"Starting Tamr dependencies.")
        _run_command(
            command=f"{tamr_install_dir}/tamr/start-dependencies.sh",
            remote_client=remote_client,
            impersonation_username=impersonation_username,
            impersonation_password=impersonation_password,
            enforce_success=True,
            verbose=verbose,
        )
    LOGGER.info(f"Starting Tamr software.")
    _run_command(
        command=f"{tamr_install_dir}/tamr/start-unify.sh",
        remote_client=remote_client,
        impersonation_username=impersonation_username,
        impersonation_password=impersonation_password,
        enforce_success=True,
        verbose=verbose,
    )


def stop_tamr(
    *,
    tamr_install_dir: str,
    include_dependencies: bool = True,
    remote_client: Optional["paramiko.SSHClient"] = None,
    impersonation_username: Optional[str] = None,
    impersonation_password: Optional[str] = None,
    verbose=False,
) -> None:
    """Stops the Tamr software and the Tamr dependencies if include_dependencies is true.

    Runs in a remote environment if an ssh client is specified otherwise runs in the local shell.
    If an impersonation_username is provided, the command is run as the provided user.
    If an impersonation_password is provided, password authentication is used for impersonation,
    otherwise sudo is used.

    Args:
        tamr_install_dir: Full path to directory where Tamr is installed
        include_dependencies: Whether Tamr dependencies should be stopped
        remote_client: An ssh client providing a remote connection
        impersonation_username: A bash user to run the command as,
            this should be the tamr install user
        impersonation_password: The password for the impersonation_username
        verbose: Whether the full command, stdout, and stderr should be logged at INFO level

    Returns:
        None

    Raises:
        RuntimeError: Raised if Tamr stop script(s) fail

    """
    LOGGER.info(f"Stopping Tamr software.")
    _run_command(
        command=f"{tamr_install_dir}/tamr/stop-unify.sh",
        remote_client=remote_client,
        impersonation_username=impersonation_username,
        impersonation_password=impersonation_password,
        enforce_success=True,
        verbose=verbose,
    )
    if include_dependencies:
        LOGGER.info(f"Stopping Tamr dependencies.")
        _run_command(
            command=f"{tamr_install_dir}/tamr/stop-dependencies.sh",
            remote_client=remote_client,
            impersonation_username=impersonation_username,
            impersonation_password=impersonation_password,
            enforce_success=True,
            verbose=verbose,
        )


def restart_tamr(
    *,
    tamr_install_dir: str,
    include_dependencies: bool = True,
    remote_client: Optional["paramiko.SSHClient"] = None,
    impersonation_username: Optional[str] = None,
    impersonation_password: Optional[str] = None,
    verbose=False,
) -> None:
    """Restarts the Tamr software and the Tamr dependencies if include_dependencies is true.

    Runs in a remote environment if an ssh client is specified otherwise runs in the local shell.
    If an impersonation_username is provided, the command is run as the provided user.
    If an impersonation_password is provided, password authentication is used for impersonation,
    otherwise sudo is used.

    Args:
        tamr_install_dir: Full path to directory where Tamr is installed
        include_dependencies: Whether Tamr dependencies should be restarted
        remote_client: An ssh client providing a remote connection
        impersonation_username: A bash user to run the command as,
            this should be the tamr install user
        impersonation_password: The password for the impersonation_username
        verbose: Whether the full command, stdout, and stderr should be logged at INFO level

    Returns:
        None

    Raises:
        RuntimeError: Raised if Tamr start or stop script(s) fail

    """
    stop_tamr(
        tamr_install_dir=tamr_install_dir,
        include_dependencies=include_dependencies,
        remote_client=remote_client,
        impersonation_username=impersonation_username,
        impersonation_password=impersonation_password,
        verbose=verbose,
    )
    start_tamr(
        tamr_install_dir=tamr_install_dir,
        include_dependencies=include_dependencies,
        remote_client=remote_client,
        impersonation_username=impersonation_username,
        impersonation_password=impersonation_password,
        verbose=verbose,
    )


def get_configs(
    *,
    config_names: Optional[List[str]] = None,
    config_search_regex: Optional[str] = None,
    user_defined_only: bool = False,
    tamr_install_dir: str,
    remote_client: Optional["paramiko.SSHClient"] = None,
    impersonation_username: Optional[str] = None,
    impersonation_password: Optional[str] = None,
) -> Dict[str, Any]:
    """Retrieves configuration values from a Tamr instance.

    Runs in a remote environment if an ssh client is specified otherwise runs in the local shell.
    If an impersonation_username is provided, the command is run as the provided user.
    If an impersonation_password is provided, password authentication is used for impersonation,
    otherwise sudo is used.

    Args:
        config_names: A list of configuration names to fetch the value for,
            when None all configurations will be fetched
        config_search_regex: A regular expression used to filter the names of the
            configurations to return
        user_defined_only: Whether to filter to only user defined config
        tamr_install_dir: Full path to directory where Tamr is installed
        remote_client: An ssh client providing a remote connection
        impersonation_username: A bash user to run the command as,
            this should be the tamr install user
        impersonation_password: The password for the impersonation_username

    Returns:
        A dictionary of Tamr configuration variables and their values
    """

    LOGGER.info(f"Retrieving {'user defined ' if user_defined_only else ''}config from Tamr.")

    user_defined_flag = " --userDefined" if user_defined_only else ""
    command = f"{tamr_install_dir}/tamr/utils/unify-admin.sh config:get{user_defined_flag}"

    if config_names is not None:
        command = f"{command} {' '.join(config_names)}"

    exit_code, stdout, stderr = _run_command(
        command=command,
        remote_client=remote_client,
        impersonation_username=impersonation_username,
        impersonation_password=impersonation_password,
        enforce_success=True,
    )

    # Tamr returns configs in yaml form, here we convert to a dictionary representation
    configs = yaml.load(stdout, Loader=yaml.SafeLoader)

    if config_search_regex is not None:
        LOGGER.info(f"Filtering to config names containing the regex '{config_search_regex}'.")
        compiled_regex = re.compile(config_search_regex, re.IGNORECASE)
        return {key: value for key, value in configs.items() if compiled_regex.search(key)}
    else:
        return configs


def get_config(
    *,
    config_name: str,
    tamr_install_dir: str,
    remote_client: Optional["paramiko.SSHClient"] = None,
    impersonation_username: Optional[str] = None,
    impersonation_password: Optional[str] = None,
) -> Any:
    """Retrieves a configuration value from a Tamr instance.

    Runs in a remote environment if an ssh client is specified otherwise runs in the local shell.
    If an impersonation_username is provided, the command is run as the provided user.
    If an impersonation_password is provided, password authentication is used for impersonation,
    otherwise sudo is used.

    Args:
        config_name: The configuration names to fetch the value for
        tamr_install_dir: Full path to directory where Tamr is installed
        remote_client: An ssh client providing a remote connection
        impersonation_username: A bash user to run the command as,
            this should be the tamr install user
        impersonation_password: The password for the impersonation_username

    Returns:
        A dictionary of Tamr configuration variables and their values
    """
    return get_configs(
        config_names=[config_name],
        config_search_regex=None,
        user_defined_only=False,
        tamr_install_dir=tamr_install_dir,
        remote_client=remote_client,
        impersonation_username=impersonation_username,
        impersonation_password=impersonation_password,
    )[config_name]


def set_configs(
    *,
    configs: Dict[str, Any],
    tamr_install_dir: str,
    remote_client: Optional["paramiko.SSHClient"] = None,
    impersonation_username: Optional[str] = None,
    impersonation_password: Optional[str] = None,
) -> Dict[str, Any]:
    """Sets configuration values in a Tamr instance.

    Runs in a remote environment if an ssh client is specified otherwise runs in the local shell.
    If an impersonation_username is provided, the command is run as the provided user.
    If an impersonation_password is provided, password authentication is used for impersonation,
    otherwise sudo is used.

    Args:
        configs: A dictionary of configuration variables and their desired values
        tamr_install_dir: Full path to directory where Tamr is installed
        remote_client: An ssh client providing a remote connection
        impersonation_username: A bash user to run the command as,
            this should be the tamr install user
        impersonation_password: The password for the impersonation_username

    Returns:
        A dictionary of Tamr configuration variables and their values for any configuration
        values that were  changed by running this command
    """
    LOGGER.info(f"Setting {len(configs.keys())} configs in Tamr.")

    # When setting values, Tamr warns about invalid config names but does not error
    # So we get all config to validate config names before sending the config:set command
    starting_config = get_configs(
        tamr_install_dir=tamr_install_dir,
        config_names=None,
        config_search_regex=None,
        user_defined_only=False,
        remote_client=remote_client,
        impersonation_username=impersonation_username,
        impersonation_password=impersonation_password,
    )

    invalid_config_names = set(configs.keys()).difference(starting_config.keys())
    if len(invalid_config_names) > 0:
        raise ValueError(
            f"Cannot set provided Tamr configs. "
            f"Invalid config names found: {invalid_config_names}"
        )

    # Run the set command
    command = (
        f"{tamr_install_dir}/tamr/utils/unify-admin.sh config:set "
        f"{' '.join([f'{key}={value}' for key, value in configs.items()])}"
    )
    _run_command(
        command=command,
        remote_client=remote_client,
        impersonation_username=impersonation_username,
        impersonation_password=impersonation_password,
        enforce_success=True,
    )

    # Retrieve all config again, to allow us to provide all config modified by the change
    # A set action can impact other config values due to formula calculations
    ending_config = get_configs(
        tamr_install_dir=tamr_install_dir,
        config_names=None,
        config_search_regex=None,
        user_defined_only=False,
        remote_client=remote_client,
        impersonation_username=impersonation_username,
        impersonation_password=impersonation_password,
    )
    return {key: value for key, value in ending_config.items() if starting_config[key] != value}


def set_config(
    *,
    config_name: str,
    config_value: Any,
    tamr_install_dir: str,
    remote_client: Optional["paramiko.SSHClient"] = None,
    impersonation_username: Optional[str] = None,
    impersonation_password: Optional[str] = None,
) -> Dict[str, Any]:
    """Sets configuration values in a Tamr instance.

    Runs in a remote environment if an ssh client is specified otherwise runs in the local shell.
    If an impersonation_username is provided, the command is run as the provided user.
    If an impersonation_password is provided, password authentication is used for impersonation,
    otherwise sudo is used.

    Args:
        config_name: The name of the configuration variable to update
        config_value: The desired value for the configuration variable
        tamr_install_dir: Full path to directory where Tamr is installed
        remote_client: An ssh client providing a remote connection
        impersonation_username: A bash user to run the command as,
            this should be the tamr install user
        impersonation_password: The password for the impersonation_username

    Returns:
        A dictionary of Tamr configuration variables and their values for any configuration
        values that were  changed by running this command
    """
    return set_configs(
        tamr_install_dir=tamr_install_dir,
        configs={config_name: config_value},
        remote_client=remote_client,
        impersonation_username=impersonation_username,
        impersonation_password=impersonation_password,
    )



"""Dataclass for validation check information"""
from dataclasses import dataclass

from tamr_toolbox.models.data_type import JsonDict


@dataclass
class ValidationCheck:
    """A dataclass for exchanging the results of validation checks.

    Args:
        passed: A boolean indicating True if the validation check passed
        details: A JSON object containing more information about the results of the validation
            check
    """

    passed: bool  # whether the check passed
    details: JsonDict  # detailed results for each check



from dataclasses import dataclass, field
from dataclasses_json import dataclass_json
from typing import List, Union
import time
import logging

from tamr_unify_client.project.resource import Project
from tamr_unify_client.operation import Operation
from tamr_toolbox.workflow.concurrent.PlanNodeStatus import PlanNodeStatus, from_plan_node
from tamr_toolbox.models.project_type import ProjectType
from tamr_toolbox.models.project_steps import (
    SchemaMappingSteps,
    CategorizationSteps,
    MasteringSteps,
    GoldenRecordsSteps,
)
from tamr_toolbox.project import schema_mapping, categorization, mastering, golden_records

LOGGER = logging.getLogger(__name__)

WORKFLOW_MAP = {
    ProjectType.SCHEMA_MAPPING_RECOMMENDATIONS: {
        SchemaMappingSteps.UPDATE_UNIFIED_DATASET: schema_mapping.jobs.update_unified_dataset
    },
    ProjectType.DEDUP: {
        MasteringSteps.UPDATE_UNIFIED_DATASET: mastering.jobs.update_unified_dataset,
        MasteringSteps.GENERATE_PAIRS: mastering.jobs.generate_pairs,
        MasteringSteps.APPLY_FEEDBACK: mastering.jobs.apply_feedback,
        MasteringSteps.UPDATE_HIGH_IMPACT_PAIRS: mastering.jobs.update_pair_predictions,
        MasteringSteps.UPDATE_CLUSTERS: mastering.jobs.update_clusters,
        MasteringSteps.PUBLISH_CLUSTERS: mastering.jobs.publish_clusters,
    },
    ProjectType.CATEGORIZATION: {
        CategorizationSteps.UPDATE_UNIFIED_DATASET: categorization.jobs.update_unified_dataset,
        CategorizationSteps.APPLY_FEEDBACK: categorization.jobs.apply_feedback,
        CategorizationSteps.UPDATE_RESULTS_ONLY: categorization.jobs.update_results_only,
    },
    ProjectType.GOLDEN_RECORDS: {
        GoldenRecordsSteps.PROFILE_GOLDEN_RECORDS: golden_records.jobs.update_profiling_info,
        GoldenRecordsSteps.UPDATE_GOLDEN_RECORDS: golden_records.jobs.update_golden_records,
        GoldenRecordsSteps.PUBLISH_GOLDEN_RECORDS: golden_records.jobs.publish_golden_records,
    },
}


@dataclass_json
@dataclass
class PlanNode:
    """
    Dataclass for the node of a Planner object
    - contains method for setting list of operations based on project type
    """

    name: str
    operations: List[Operation]
    project: Project
    priority: int
    current_op: Operation
    status: PlanNodeStatus = PlanNodeStatus.PLANNED
    train: bool = False
    project_type: ProjectType = field(init=False)
    project_steps: Union[
        List[SchemaMappingSteps],
        List[MasteringSteps],
        List[GoldenRecordsSteps],
        List[CategorizationSteps],
    ] = field(init=False)
    current_step: Union[
        SchemaMappingSteps, MasteringSteps, CategorizationSteps, GoldenRecordsSteps
    ] = None
    steps_to_run: Union[
        List[SchemaMappingSteps],
        List[MasteringSteps],
        List[GoldenRecordsSteps],
        List[CategorizationSteps],
    ] = None

    def __post_init__(self):
        # set the project type
        self.project_type = ProjectType[self.project.type]

        # then set project steps based on project type
        # first for schema mapping projects
        if self.project_type == ProjectType.SCHEMA_MAPPING_RECOMMENDATIONS:
            self.project_steps = [SchemaMappingSteps.UPDATE_UNIFIED_DATASET]
        # now for mastering projects
        elif self.project_type == ProjectType.DEDUP:
            if self.train:
                self.project_steps = [
                    MasteringSteps.UPDATE_UNIFIED_DATASET,
                    MasteringSteps.GENERATE_PAIRS,
                    MasteringSteps.APPLY_FEEDBACK,
                    MasteringSteps.UPDATE_HIGH_IMPACT_PAIRS,
                    MasteringSteps.UPDATE_CLUSTERS,
                    MasteringSteps.PUBLISH_CLUSTERS,
                ]
            else:
                self.project_steps = [
                    MasteringSteps.UPDATE_UNIFIED_DATASET,
                    MasteringSteps.GENERATE_PAIRS,
                    MasteringSteps.UPDATE_HIGH_IMPACT_PAIRS,
                    MasteringSteps.UPDATE_CLUSTERS,
                    MasteringSteps.PUBLISH_CLUSTERS,
                ]
        # now for categorization projects
        elif self.project_type == ProjectType.CATEGORIZATION:
            if self.train:
                self.project_steps = [
                    CategorizationSteps.UPDATE_UNIFIED_DATASET,
                    CategorizationSteps.APPLY_FEEDBACK,
                    CategorizationSteps.UPDATE_RESULTS_ONLY,
                ]
            else:
                self.project_steps = [
                    CategorizationSteps.UPDATE_UNIFIED_DATASET,
                    CategorizationSteps.UPDATE_RESULTS_ONLY,
                ]
        # finally for golden record projects
        elif self.project_type == ProjectType.GOLDEN_RECORDS:
            self.project_steps = [
                GoldenRecordsSteps.PROFILE_GOLDEN_RECORDS,
                GoldenRecordsSteps.UPDATE_GOLDEN_RECORDS,
                GoldenRecordsSteps.PUBLISH_GOLDEN_RECORDS,
            ]

        else:
            raise NotImplementedError(
                f"Don't know how to create a PlanNode object for project type {self.project_type}"
            )


def poll(plan_node: PlanNode) -> PlanNode:
    """
    Polls the status of the current_op object and returns a new PlanNode with updated status

    Args:
        plan_node: the PlanNode object to poll

    Returns:
        Copy of the original plan node with status updated based on the status of the current_op
    """
    # get current op and see if it is None (i.e. if plannode hasn't been triggered)
    current_op = plan_node.current_op

    if current_op is None:
        # if this node hasn't been triggered just return the current status in case
        # it has been updated by upstream actions
        updated_plan_node_status = plan_node.status
        updated_current_op = None
        updated_operations = None
    else:
        # current op exists so update it
        updated_current_op = plan_node.current_op.poll()
        # also update the list of all ops - rigamarole is to avoid updating current op twice
        # since it already exists in plan_node.operations list
        updated_operations = [x.poll() for x in plan_node.operations if x != current_op]
        updated_operations.append(updated_current_op)

        # if the op status changed set the plan node's current op to the updated one
        # and use from_plan_node to capture logic around in progress
        if updated_current_op.state != current_op.state:
            # update the current op
            plan_node.current_op = updated_current_op
            plan_node.operations = updated_operations
            updated_plan_node_status = from_plan_node(plan_node)
        # if the op didn't change status neither will the node so just pass through
        else:
            updated_plan_node_status = plan_node.status

    return PlanNode(
        name=plan_node.name,
        operations=updated_operations,
        project=plan_node.project,
        priority=plan_node.priority,
        current_op=updated_current_op,
        status=updated_plan_node_status,
        current_step=plan_node.current_step,
        steps_to_run=plan_node.steps_to_run,
    )


def run_next_step(plan_node: PlanNode) -> PlanNode:
    """
    Takes a plan node and runs the project

    Args:
        plan_node: the node to run

    Returns:
        updated plan node
    """
    # if current_step is None this node has never been run,
    # so set it and the steps_to_run from project_steps
    current_step = plan_node.current_step
    if current_step is None:
        steps_to_run = plan_node.project_steps[1:]
        next_step = plan_node.project_steps[0]
    # else we are on step 2+ so user and trim list of steps_to_run
    else:
        LOGGER.debug(f"plan node in trigger next step: {plan_node}")
        next_step = plan_node.steps_to_run[0]
        steps_to_run = plan_node.steps_to_run[1:]

    LOGGER.info(f"running step {next_step.value} for project {plan_node.name}")
    # We only call methods that return a list with one and only one operation
    current_op = WORKFLOW_MAP[plan_node.project_type][next_step](
        plan_node.project, process_asynchronously=True
    )[0]
    # handle case where operations is empty list (nothing has been run)
    if plan_node.operations is None:
        operations_list = [current_op]
    else:
        operations_list = [x for x in plan_node.operations]
        if current_op not in operations_list:
            operations_list.append(current_op)
    # don't forget to update the status - and first update the current op and operations list
    plan_node.current_op = current_op
    plan_node.operations = operations_list
    plan_node.current_step = next_step
    status = from_plan_node(plan_node)
    return PlanNode(
        name=plan_node.name,
        operations=operations_list,
        current_op=current_op,
        priority=plan_node.priority,
        project=plan_node.project,
        steps_to_run=steps_to_run,
        status=status,
        current_step=next_step,
    )


def monitor(
    nodes: List[PlanNode], *, timeout: int = 12, polling_interval: int = 30
) -> List[PlanNode]:
    """
    Monitors the status of a list of PlanNodes, when one on that list changes to
        either failed/succeeded/cancelled it returns the list

    Args:
        nodes: list of nodes to monitor
        timeout: number of hours to poll before timing out for change in job status
        polling_interval: the amount of time in seconds to wait between polling

    Returns:
        the same list of planNodes with updated statuses
    """
    # if empty then return immediate and log a warning
    if not nodes:
        LOGGER.warning("Jobs list is empty! Something is probably wrong.")
        return []

    start_time = time.time()
    timeout_in_seconds = 3600 * timeout

    # get initial statuses
    initial_statuses = {x.name: x.status for x in nodes}

    # get project names for logging
    running_projects = "\n".join([x.name for x in nodes])
    LOGGER.info(
        f"starting to monitor projects:\n {running_projects}\n with timeout of {timeout} hours"
    )
    # poll only as long as we haven't gotten past timeout
    while (time.time() - start_time) < timeout_in_seconds:

        # sleep at beginning of loop for polling interval
        time.sleep(polling_interval)

        # poll the jobs and get updated nodes
        updated_nodes = [poll(x) for x in nodes]
        # generate updated statuses
        updated_statuses = {x.name: x.status for x in updated_nodes}

        # if status has changed return list
        if initial_statuses != updated_statuses:
            return updated_nodes

    # if we got here it timed out and raise runtime error
    error_message = f"Monitoring jobs:\n {running_projects}\n timed out after {timeout} hours!"
    LOGGER.error(error_message)
    raise RuntimeError(error_message)



"""Dataclasses, enums, and  types related to Tamr"""
from . import project_type
from . import attribute_type
from . import data_type
from . import operation_state
from . import validation_check
from . import project_artifacts
from . import project_steps

__all__ = [
    "project_type",
    "attribute_type",
    "data_type",
    "operation_state",
    "validation_check",
    "project_artifacts",
    "project_steps",
]



"""Tasks related to the version of Tamr instances"""
from functools import wraps
import inspect
import json
import logging
import warnings
from collections.abc import Callable
from typing import List, Optional

from packaging.version import parse
from tamr_unify_client import Client

LOGGER = logging.getLogger(__name__)
logging.captureWarnings(True)


def current(client: Client) -> str:
    """
    Gets the version of Tamr for provided client

    Args:
        client: Tamr client

    Returns:
        String representation of Tamr version
    """
    if client.port == 9100:
        url = "/api/versioned/service/version"
    else:
        url = "/api/service/version"
    response = client.get(url).successful()
    return json.loads(response.content)["version"]


def _as_float(version: str) -> float:
    """
    Converts string Tamr version to an orderable numeric representation

    The numeric representation is designed to allow for the ordering of Tamr versions from oldest
    to newest. The values are not guaranteed to be sequential even if the versions are sequential.
    Formats covered:
        year.release.patch (2020.001.1)
        major.minor.patch (0.43.0)

    Args:
        version: String representation of Tamr version

    Returns:
        Numeric representation of Tamr version
    """
    warnings.warn("Use `packaging.version.parse() instead.'", DeprecationWarning)
    version_split = version.split(".")

    if len(version_split) != 3:
        raise ValueError(f"Tamr version {version} does not match known patterns.")

    version_split = [float(x) for x in version_split]
    return (version_split[0] * 1000) + version_split[1] + (version_split[2] / 10)


def _get_tamr_versions_from_function_args(*args, **kwargs) -> List[str]:
    """
    Gets the Tamr version of any/all relevant inputs

    Args:
        *args: Any argument that may/may not be linkable to a versioned Tamr client
        **kwargs: Any argument that may/may not be linkable to a versioned Tamr client

    Returns:
        List of all Tamr versions inputted to the function

    """
    all_args = locals()
    args = [arg for arg in all_args["args"]]
    kwargs = list(all_args["kwargs"].values())
    all_args_parsed = args + kwargs
    response = []

    # Return the client version (if we can find it)
    for arg in all_args_parsed:
        if type(arg) is Client:
            response.append(current(arg))
        elif hasattr(arg, "client"):
            if type(arg.client) is Client:
                response.append(current(arg.client))

    return response


def is_version_condition_met(
    *,
    tamr_version: str,
    min_version: str,
    max_version: Optional[str] = None,
    exact_version: bool = False,
    raise_error: bool = False,
) -> bool:
    """
    Check if Tamr version is valid.

    Args:
        tamr_version:
            The version of Tamr being considered
        min_version:
            The earliest version of Tamr
        max_version:
            The latest version of Tamr.
            Default None, in which case no max version is tested for.
        exact_version:
            Compare against only one release of Tamr. Default is False
        raise_error:
            If True, raise an error if the version condition is not met. Default is False.

    Raises:
        ValueError: if `min_version` is greater than `max_version`
        EnvironmentError: if `raise_error` is True, and the condition is not met

    Notes:
        Patch versions (major.minor.patch) are excluded from the comparison
        If exact_version is True, max_version will be ignored

    See Also:
        utils.version.func_requires_tamr_version
    """

    error_str = None

    tamr_version_sub = parse(tamr_version).release[:2]
    min_version_sub = parse(min_version).release[:2]

    if exact_version:
        if not min_version_sub == tamr_version_sub:
            error_str = f"must be exactly {min_version}."

    elif max_version:
        max_version_sub = parse(max_version).release[:2]
        if min_version_sub > max_version_sub:
            raise ValueError("min_version must be smaller than max_version")

        if not min_version_sub <= tamr_version_sub <= max_version_sub:
            error_str = f"must be between {min_version} and {max_version}."

    elif not min_version_sub <= tamr_version_sub:
        error_str = f"must be at least {min_version}."

    if error_str and raise_error:
        raise EnvironmentError(f"Using Tamr version(s) {tamr_version}, but " + error_str)
    else:
        return not error_str


def requires_tamr_version(
    min_version: str, max_version: Optional[str] = None, exact_version: bool = False
) -> Callable:
    """
    Pie decorator for Tamr version checking

    Args:
        min_version:
            The earliest version of Tamr that supports the function
        max_version:
            The latest version of Tamr that supports the function.
            Default None, supporting all latest releases of Tamr
        exact_version:
            If True, only support one release of Tamr. Default is False

    Examples:
        >>> @requires_tamr_version(min_version="2021.002")
        >>> def refresh_dataset(tamr_dataset, *args, **kwargs):
        >>>     return tamr_dataset.refresh()

    Raises:
        ValueError: if `min_version` is greater than `max_version`
        EnvironmentError: if `raise_error` is True, and the condition is not met

    Notes:
        This decorator only inspects the Tamr version of arguments going into the
        function, and not new instances of Tamr referred to within functional code

        Patch versions (major.minor.patch) are excluded from the comparison

    See Also:
        utils.version.is_version_condition_met
    """

    def _decorator(func):
        @wraps(func)
        def _inspector(*args, **kwargs):
            for tamr_version in _get_tamr_versions_from_function_args(*args, **kwargs):
                is_version_condition_met(
                    tamr_version=tamr_version,
                    min_version=min_version,
                    max_version=max_version,
                    exact_version=exact_version,
                    raise_error=True,
                )

            return func(*args, **kwargs)

        return _inspector

    return _decorator


def enforce_after_or_equal(client: Client, *, compare_version: str) -> None:
    """Raises an exception if the version of the Tamr client is before the provided compare version
        Will be deprecated in favour of raise_warn_tamr_version()

    Args:
        client: Tamr client
        compare_version: String representation of Tamr version

    Returns:
        None

    See Also:
        raise_warn_tamr_version
        ensure_tamr_version
    """
    warnings.warn("Use `is_version_condition_met'", DeprecationWarning)

    current_version = current(client)
    if _as_float(current_version) < _as_float(compare_version):
        raise NotImplementedError(
            f"This function is not available in Tamr {current_version}. "
            f"Upgrade to Tamr {compare_version} or later to use this function."
        )


def _deprecated_warning(func: Callable, *, message: str) -> Callable:
    """
    Decorator to log a warning message when the passed function is called.
    Intended for warning about deprecated functions.

    Args:
        func: The function to attach the warning message to
        message: The warning message
    Returns:
        The decorated function
    """
    warnings.warn(
        "Use warnings.warn with `DeprecationWarning', ensuring"
        "`logging.captureWarnings' is True",
        DeprecationWarning,
    )

    def warning(*args, **kwargs):
        try:
            current_frame = inspect.currentframe()
            previous_frame = current_frame.f_back
            calling_lineno = previous_frame.f_lineno
            calling_func = previous_frame.f_code
            LOGGER.warning(
                f"In {calling_func.co_filename}:{calling_func.co_name}:{calling_lineno} {message}"
            )
        finally:
            # Avoid reference cycle
            # https://docs.python.org/3/library/inspect.html#the-interpreter-stack
            del current_frame
            del previous_frame
        return func(*args, **kwargs)

    return warning



"""Tasks related to connecting to a Tamr instance"""
import logging
import re
from base64 import b64decode

import requests
from json import dumps
from typing import Optional, Union
from time import sleep, time as now

from requests import Response
from tamr_unify_client import Client
from tamr_unify_client.auth import UsernamePasswordAuth
from tamr_unify_client.auth import JwtTokenAuth

from tamr_toolbox.utils.version import requires_tamr_version

LOGGER = logging.getLogger(__name__)

TAMR_JWT_RELEASE_VERSION = "2022.010.0"


def health_check(client: Client) -> bool:
    """
    Query the health check API and check if each service is healthy (returns True)

    Args:
        client: the tamr client

    Returns:
        True if all services are healthy, False if unhealthy
    """
    try:
        response = client.get(endpoint="/api/service/health")
        healthy_status = all([value["healthy"] for value in response.json().values()])
        if healthy_status:
            LOGGER.info(f"Client is healthy: {dumps(response.json(), indent=2)}")
        else:
            LOGGER.error(f"Client is unhealthy: {dumps(response.json(), indent=2)}")
        return healthy_status

    except requests.exceptions.ConnectionError as e:
        LOGGER.error(f"Could not connect to {client.host}. Error: {e}")
        return False


def create(
    *,
    username: str,
    password: str,
    host: str,
    port: Optional[Union[str, int]] = 9100,
    protocol: str = "http",
    base_path: str = "/api/versioned/v1/",
    session: Optional[requests.Session] = None,
    store_auth_cookie: bool = False,
    enforce_healthy: bool = False,
) -> Client:
    """Creates a Tamr client from the provided configuration values

    Args:
        username: The username to log access Tamr as
        password: the password for the user
        host: The ip address of Tamr
        port: The port of the Tamr UI. Pass a value of `None` to specify an address with no port
        protocol: https or http
        base_path: Optional argument to specify a different base path
        session: Optional argument to pass an existing requests Session
        store_auth_cookie: If true will allow Tamr authentication cookie to be stored and reused
        enforce_healthy: If true will enforce a healthy state upon creation

    Returns:
        Tamr client
    """
    full_address = f"{protocol}://{host}:{port}" if port is not None else f"{protocol}://{host}"
    LOGGER.info(f"Creating client as user {username} at {full_address}.")
    client = Client(
        auth=UsernamePasswordAuth(username=username, password=password),
        host=host,
        port=int(port) if port is not None else None,
        protocol=protocol,
        base_path=base_path,
        session=session,
        store_auth_cookie=store_auth_cookie,
    )
    if enforce_healthy:
        if not health_check(client):
            LOGGER.error(f"Tamr is not healthy. Check logs and Tamr.")
            raise SystemError("Tamr is not healthy. Check logs and Tamr.")
    return client


@requires_tamr_version(min_version=TAMR_JWT_RELEASE_VERSION)
def create_with_jwt(
    *,
    token: str,
    host: str,
    port: Optional[Union[str, int]] = 9100,
    protocol: str = "http",
    base_path: str = "/api/versioned/v1/",
    session: Optional[requests.Session] = None,
    store_auth_cookie: bool = False,
    enforce_healthy: bool = False,
) -> Client:
    """Creates a Tamr client from the provided configuration values using a JWT token instead
    of a username and password. Note that this feature is only available on v2022.010.0 or later.

    Args:
        token: A JWT token to authenticate the client
        host: The ip address of Tamr
        port: The port of the Tamr UI. Pass a value of `None` to specify an address with no port
        protocol: https or http
        base_path: Optional argument to specify a different base path
        session: Optional argument to pass an existing requests Session
        store_auth_cookie: If true will allow Tamr authentication cookie to be stored and reused
        enforce_healthy: If true will enforce a healthy state upon creation

    Returns:
        Tamr client
    """
    full_address = f"{protocol}://{host}:{port}" if port is not None else f"{protocol}://{host}"
    LOGGER.info(f"Creating client using JWT token at {full_address}.")
    client = Client(
        auth=JwtTokenAuth(token=token),
        host=host,
        port=int(port) if port is not None else None,
        protocol=protocol,
        base_path=base_path,
        session=session,
        store_auth_cookie=store_auth_cookie,
    )

    if enforce_healthy:
        if not health_check(client):
            LOGGER.error(f"Tamr is not healthy. Check logs and Tamr.")
            raise SystemError("Tamr is not healthy. Check logs and Tamr.")
    return client


def get_with_connection_retry(
    client: Client, api_endpoint: str, *, timeout_seconds: int = 600, sleep_seconds: int = 20
) -> requests.Response:
    """Will handle exceptions when attempting to connect to the Tamr API.
        This is used to handle connection issues when Tamr restarts due to a restore.

    Args:
        client: A Tamr client object
        api_endpoint: Tamr API endpoint
        timeout_seconds: Amount of time before a timeout error is thrown. Default is 600 seconds
        sleep_seconds: Amount of time in between attempts to connect to Tamr.

    Returns:
        A response object from API request."""
    started = now()
    while timeout_seconds is None or now() - started < timeout_seconds:
        try:
            response = client.get(api_endpoint)
            return response
        except requests.exceptions.ConnectionError as e:
            # If we got for example a connection refused exception, try again later
            LOGGER.warning(f"Caught exception in connect {e}")
            sleep(sleep_seconds)
    raise TimeoutError(f"Took longer than {timeout_seconds} seconds to connect to tamr.")


def poll_endpoint(
    client: Client,
    api_endpoint: str,
    *,
    poll_interval_seconds: int = 3,
    polling_timeout_seconds: Optional[int] = None,
    connection_retry_timeout_seconds: int = 600,
) -> requests.Response:
    """Waits until job has a state of Canceled, Succeeded, or Failed.

    Args:
        client: A Tamr client object
        api_endpoint: Tamr API endpoint
        poll_interval_seconds: Amount of time in between polls of job state.
        polling_timeout_seconds: Amount of time before a timeout error is thrown.
        connection_retry_timeout_seconds: Amount of time before timeout error is thrown
            during connection retry.

    Returns:
        A response object from API request.
    """

    started = now()
    op = get_with_connection_retry(
        client=client,
        api_endpoint=api_endpoint,
        timeout_seconds=connection_retry_timeout_seconds,
        sleep_seconds=poll_interval_seconds,
    )
    state = op.json()["state"]
    while polling_timeout_seconds is None or now() - started < polling_timeout_seconds:
        if state in ["PENDING", "RUNNING"]:
            sleep(poll_interval_seconds)
        elif state in ["CANCELED", "SUCCEEDED", "FAILED"]:
            return op
        op = get_with_connection_retry(
            client=client,
            api_endpoint=api_endpoint,
            timeout_seconds=connection_retry_timeout_seconds,
            sleep_seconds=poll_interval_seconds,
        )
        state = op.json()["state"]
    raise TimeoutError(f"Took longer than {polling_timeout_seconds} seconds to connect to tamr.")


def _from_response(response: Response) -> Client:
    """Creates a Tamr Client based on a previous api response

    Args:
        response: The response to base the Client on

    Returns:
        New Tamr Client based on the previous response
    """
    request = response.request
    url_matcher = re.match(r"(https?)://(.*):(\d{4})(.*)", request.url)

    auth_hash_matcher = re.match(r"BasicCreds (.*)", request.headers.get("Authorization"))
    creds_matcher = re.match(r"(.*):(.*)", b64decode(auth_hash_matcher.group(1)).decode("latin1"))

    return create(
        username=creds_matcher.group(1),
        password=creds_matcher.group(2),
        host=url_matcher.group(2),
        port=url_matcher.group(3),
        protocol=url_matcher.group(1),
    )



""" Project artifacts data classes """
from dataclasses import dataclass


@dataclass()
class SchemaMappingArtifacts:
    """A dataclass representing artifact codes for Schema Mapping projects in Tamr

    Args:
        UNIFIED_ATTRIBUTES: artifact name for a schema mapping project
        TRANSFORMATIONS: artifact name for a schema mapping project
        SMR_MODEL: artifact name for a schema mapping project
        RECORD_COMMENTS: artifact name for a schema mapping project
    """

    # Schema Mapping artifacts
    UNIFIED_ATTRIBUTES: str = "UNIFIED_ATTRIBUTES"
    TRANSFORMATIONS: str = "TRANSFORMATIONS"
    SMR_MODEL: str = "SMR_MODEL"
    RECORD_COMMENTS: str = "RECORD_COMMENTS"


@dataclass()
class MasteringArtifacts:
    """A dataclass representing artifact codes for Mastering projects in Tamr

    Args:
        UNIFIED_ATTRIBUTES: artifact name for a Mastering project
        TRANSFORMATIONS: artifact name for a Mastering project
        SMR_MODEL: artifact name for a Mastering project
        RECORD_COMMENTS: artifact name for a Mastering project

        MASTERING_CONFIGURATION: artifact name for a Mastering project
        USER_DEFINED_SIGNALS: artifact name for a Mastering project
        MASTERING_FUNCTIONS: artifact name for a Mastering project
        RECORD_PAIR_COMMENTS: artifact name for a Mastering project
        RECORD_PAIR_VERIFIED_LABELS: artifact name for a Mastering project
        RECORD_PAIR_UNVERIFIED_LABELS: artifact name for a Mastering project
        RECORD_PAIR_ASSIGNMENTS: artifact name for a Mastering project
        CLUSTERING_MODEL: artifact name for a Mastering project
        PUBLISHED_CLUSTERS: artifact name for a Mastering project
        CLUSTER_RECORD_VERIFICATIONS: artifact name for a Mastering project
        CLUSTER_ASSIGNMENTS: artifact name for a Mastering project
    """

    # Schema Mapping artifacts
    UNIFIED_ATTRIBUTES: str = "UNIFIED_ATTRIBUTES"
    TRANSFORMATIONS: str = "TRANSFORMATIONS"
    SMR_MODEL: str = "SMR_MODEL"
    RECORD_COMMENTS: str = "RECORD_COMMENTS"
    # Mastering artifacts
    MASTERING_CONFIGURATION: str = "MASTERING_CONFIGURATION"
    USER_DEFINED_SIGNALS: str = "USER_DEFINED_SIGNALS"
    MASTERING_FUNCTIONS: str = "MASTERING_FUNCTIONS"
    RECORD_PAIR_COMMENTS: str = "RECORD_PAIR_COMMENTS"
    RECORD_PAIR_VERIFIED_LABELS: str = "RECORD_PAIR_VERIFIED_LABELS"
    RECORD_PAIR_UNVERIFIED_LABELS: str = "RECORD_PAIR_UNVERIFIED_LABELS"
    RECORD_PAIR_ASSIGNMENTS: str = "RECORD_PAIR_ASSIGNMENTS"
    CLUSTERING_MODEL: str = "CLUSTERING_MODEL"
    PUBLISHED_CLUSTERS: str = "PUBLISHED_CLUSTERS"
    CLUSTER_RECORD_VERIFICATIONS: str = "CLUSTER_RECORD_VERIFICATIONS"
    CLUSTER_ASSIGNMENTS: str = "CLUSTER_ASSIGNMENTS"


@dataclass()
class CategorizationArtifacts:
    """A dataclass representing artifact codes for Categorization projects in Tamr

    Args:
        UNIFIED_ATTRIBUTES: artifact name for a Categorization project
        TRANSFORMATIONS: artifact name for a Categorization project
        SMR_MODEL: artifact name for a Categorization project
        RECORD_COMMENTS: artifact name for a Categorization project

        CATEGORIZATION_CONFIGURATION: artifact name for a Categorization project
        CATEGORIZATION_FUNCTIONS: artifact name for a Categorization project
        CATEGORIZATION_VERIFIED_LABELS: artifact name for a Categorization project
        CATEGORIZATION_TAXONOMIES: artifact name for a Categorization project
        CATEGORIZATION_MODEL: artifact name for a Categorization project
        CATEGORIZATION_FEEDBACK: artifact name for a Categorization project
    """

    # Schema Mapping artifacts
    UNIFIED_ATTRIBUTES: str = "UNIFIED_ATTRIBUTES"
    TRANSFORMATIONS: str = "TRANSFORMATIONS"
    SMR_MODEL: str = "SMR_MODEL"
    RECORD_COMMENTS: str = "RECORD_COMMENTS"
    # Categorization artifacts
    CATEGORIZATION_CONFIGURATION: str = "CATEGORIZATION_CONFIGURATION"
    CATEGORIZATION_FUNCTIONS: str = "CATEGORIZATION_FUNCTIONS"
    CATEGORIZATION_VERIFIED_LABELS: str = "CATEGORIZATION_VERIFIED_LABELS"
    CATEGORIZATION_TAXONOMIES: str = "CATEGORIZATION_TAXONOMIES"
    CATEGORIZATION_MODEL: str = "CATEGORIZATION_MODEL"
    CATEGORIZATION_FEEDBACK: str = "CATEGORIZATION_FEEDBACK"


@dataclass()
class GoldenRecordsArtifacts:
    """A dataclass representing artifact codes for Golden Records projects in Tamr

    GR_CONFIGURATION: artifact name for a Golden Records project
    GR_RULES: artifact name for a Golden Records project
    GR_OVERRIDES: artifact name for a Golden Records project
    """

    # Golden Records artifacts
    GR_CONFIGURATION: str = "GR_CONFIGURATION"
    GR_RULES: str = "GR_RULES"
    GR_OVERRIDES: str = "GR_OVERRIDES"


@dataclass()
class ProjectArtifacts:
    """A dataclass representing the project artifact codes in Tamr

    Args:
       SCHEMA_MAPPING: SchemaMappingArtifacts dataclass instance
       MASTERING: MasteringArtifacts dataclass instance
       CATEGORIZATION: MasteringArtifacts dataclass instance
       GOLDEN_RECORDS: MasteringArtifacts dataclass instance
    """

    SCHEMA_MAPPING = SchemaMappingArtifacts()
    MASTERING = MasteringArtifacts()
    CATEGORIZATION = CategorizationArtifacts()
    GOLDEN_RECORDS = GoldenRecordsArtifacts()



"""Tasks related to running jobs for groups of Tamr projects"""
import time
from typing import List, Optional
import logging

from tamr_unify_client.operation import Operation
from tamr_unify_client.project.resource import Project, Dataset
from tamr_toolbox.models.project_type import ProjectType

from tamr_toolbox.project import mastering, categorization, golden_records, schema_mapping
from requests import HTTPError

LOGGER = logging.getLogger(__name__)


def run(
    project_list: List[Project],
    *,
    run_apply_feedback: bool = False,
    run_estimate_pair_counts: bool = False,
    run_profile_unified_datasets: bool = False,
    sleep_interval: int = 0,
) -> List[Operation]:
    """Run multiple projects in order

    Args:
        project_list: A list of Tamr projects
        run_apply_feedback: Whether train should be called on the pair matching model
            or categorization model (based on project type)
        run_estimate_pair_counts: Whether an estimate pairs job should be run
        run_profile_unified_datasets: Whether unified datasets should be re-profiled
        sleep_interval: Number of seconds to sleep between job submissions.
            Useful in some pipeline situations

    Returns:
        The operations that were run

    Raises:
        NotImplementedError: Raised if run() is called on an unsupported project type

    """
    operations = []
    for project in project_list:
        target_type = ProjectType[project.type]
        if target_type == ProjectType.SCHEMA_MAPPING_RECOMMENDATIONS:
            operations.extend(schema_mapping.jobs.run(project))
        elif target_type == ProjectType.CATEGORIZATION:
            operations.extend(
                categorization.jobs.run(
                    project.as_categorization(), run_apply_feedback=run_apply_feedback
                )
            )
        elif target_type == ProjectType.DEDUP:
            operations.extend(
                mastering.jobs.run(
                    project.as_mastering(),
                    run_apply_feedback=run_apply_feedback,
                    run_estimate_pair_counts=run_estimate_pair_counts,
                )
            )
        elif target_type == ProjectType.GOLDEN_RECORDS:
            operations.extend(golden_records.jobs.run(project))
        else:
            error_msg = (
                f"Function run.continuous_operations not defined for project type {project.type}"
            )
            LOGGER.error(error_msg)
            raise NotImplementedError(error_msg)

        # Excluding GOLDEN_RECORDS as profile information is refreshed as part of job:
        # breakpoint()
        if (
            (target_type == ProjectType.SCHEMA_MAPPING_RECOMMENDATIONS)
            | (target_type == ProjectType.CATEGORIZATION)
            | (target_type == ProjectType.DEDUP)
        ):
            if run_profile_unified_datasets:
                LOGGER.info(f"Checking for profile for {project.unified_dataset().name}")
                try:
                    profile = project.unified_dataset().profile()
                except HTTPError:
                    LOGGER.info(
                        f"Profile not found for {project.unified_dataset().name}. Creating"
                        f" a profile."
                    )
                    project.unified_dataset().create_profile()
                    profile = project.unified_dataset().profile()
                LOGGER.info(f"Refreshing profile for {project.unified_dataset().name}")
                operations.append(profile.refresh())

        # if we need to sleep before the next iteration then do so
        # default is zero so don't need an if statement
        time.sleep(sleep_interval)

    return operations


def _find_upstream_projects(
    project: Project,
    all_projects: Optional[List[Project]] = None,
    project_list: Optional[List[str]] = None,
    upstream_projects: Optional[List[str]] = None,
) -> List[Project]:
    """Check for upstream projects that have source datasets that require
    updating
    Args:
        project: a tamr project
        track of projects that need to be checked for upstream dependencies
        all_projects: list of all projects that exist within Tamr
        project_list: keeps track of projects that have yet to be checked
        for upstream dependencies.
        upstream_projects: list to keep track of projects within a chain of projects
    Returns:
        upstream_projects: final list of tamr projects in a chained
        workflow
    """
    # Obtain the name of project to be updated and initiate the list of
    # projects (project_list) that are maintained within the check_for_upstream_project
    # function to check for upstream dependencies

    # Get project_id from project
    project_id = project.resource_id

    if not upstream_projects:
        upstream_projects = []
        project_list = [project_id]

        all_projects = {}

        for project in project.client.projects:
            all_projects[project.name] = project.resource_id

    for project_id in project_list:
        # Obtain resource id of project to be updated from the dictionary of
        # project details
        project = project.client.projects.by_resource_id(project_id)
        if project.resource_id not in upstream_projects:
            upstream_projects.append(project.resource_id)

        # Remove the current project from the list tracking projects that
        # need to be reviewed for upstream projects/dependencies
        project_list.remove(project_id)

        project_source_datasets = project.client.get(
            f"/api/versioned/v1/projects/{project.resource_id}/inputDatasets"
        ).json()
        relative_ids = [dataset["relativeId"] for dataset in project_source_datasets]

        # Check which upstream projects the upstream datasets derive from
        for relative_id in relative_ids:
            dataset_details = project.client.datasets.by_relative_id(relative_id)

            # If these projects are linked to further datasets upstream,
            # keep track of these upstream projects in project_list
            upstream_datasets = dataset_details.usage().usage.output_from_project_steps
            if upstream_datasets:
                for upstream_dataset in upstream_datasets:
                    upstream_project = upstream_dataset.project_name
                    project_id = all_projects[upstream_project]
                    if project_id not in project_list:
                        project_list.append(project_id)

    # If there are projects added to the project list, check whether these
    # projects have dependencies upstream by rerunning the
    # check_for_upstream_datasets recursively
    if project_list:
        return _find_upstream_projects(
            project=project,
            project_list=project_list,
            all_projects=all_projects,
            upstream_projects=upstream_projects,
        )
    else:
        # If the first project in the chain is reached, return the list of
        # projects that need to be updated

        # Reverse list of projects so that the first project in the chain is
        # returned first
        upstream_projects.reverse()

        upstream_projects = [
            project.client.projects.by_resource_id(project_id) for project_id in upstream_projects
        ]

        return upstream_projects


def get_upstream_projects(project: Project) -> List[Project]:
    """Check for upstream projects associated with a specified project

    Args:
        project: the tamr project for which associated upstream projects are retrieved
    """

    # Retrieve upstream projects (if any) from a specified Project
    upstream_projects = _find_upstream_projects(project)

    return upstream_projects


def get_project_output_datasets(project: Project) -> List[Dataset]:
    """Retrieves datasets produced by a given Tamr project

    Args:
        project: the Tamr project for which associated output datasets are retrieved

    Returns:
        The list of Tamr datasets output from the project"""
    input_datasets_dependencies = [
        dep
        for ds in {input_ds for input_ds in project.input_datasets()}
        for dep in ds.usage().dependencies
    ]
    output_dataset_names = {
        dep.dataset_name
        for dep in input_datasets_dependencies
        if dep.output_from_project_steps
        and project.name == dep.output_from_project_steps[0].project_name
    }

    return [ds for ds in project.client.datasets if ds.name in output_dataset_names]



"""Tasks related to Tamr Projects"""
from . import schema_mapping
from . import mastering
from . import golden_records
from . import categorization
from ._common import export_artifacts, import_artifacts

__all__ = [
    "schema_mapping",
    "mastering",
    "golden_records",
    "categorization",
    "export_artifacts",
    "import_artifacts",
]



"""Tasks related to Tamr operations (or jobs)"""
import logging
from typing import Union, List, Optional
from time import sleep, time as now

from tamr_unify_client import Client
from tamr_unify_client.operation import Operation
from tamr_toolbox.models.data_type import JsonDict
from tamr_toolbox.models.operation_state import OperationState

LOGGER = logging.getLogger(__name__)


def enforce_success(operation: Operation) -> None:
    """Raises an error if an operation fails

    Args:
        operation: A Tamr operation
    """
    if not operation.succeeded():
        raise RuntimeError(
            f"Operation {operation.resource_id} failed. Description: {operation.description}."
            f"Status: {operation.status}"
        )


def from_resource_id(tamr: Client, *, job_id: Union[int, str]) -> Operation:
    """Create an operation from a job id

    Args:
        tamr: A Tamr client
        job_id: A job ID

    Returns:
        A Tamr operation
    """
    # Tamr sometimes returns a job id of -1 if everthing is up-to-date so the requested task
    # would result in no change -- make a dummy operation in this case
    if str(job_id) == "-1":
        return Operation.from_json(tamr, _dummy_no_op_response())

    return Operation.from_resource_id(tamr, str(job_id))


def _dummy_no_op_response(code: str = "job ID -1") -> JsonDict:
    """
    Create a dict to represent 'no-operation operation', in cases where all results that would be
    produced by an operation are already up-to-date so there's nothing to do.

    Args:
        code: response code include in description, Default 'job ID -1'

    Returns:
        dict from which dummy operation object can be constructed
    """

    _never = "0000-00-00T00:00:00.000Z"
    _description = f"""Tamr returned {code} for this operation, indicating that all
        results that would be produced by the operation are already up-to-date."""
    dummy_dict = {
        "id": "-1",
        "type": "NOOP",
        "description": _description,
        "status": {"state": "SUCCEEDED", "startTime": _never, "endTime": _never, "message": ""},
        "created": {"username": "", "time": _never, "version": "-1"},
        "lastModified": {"username": "", "time": _never, "version": "-1"},
        "relativeId": "operations/-1",
    }
    return dummy_dict


def get_latest(tamr: Client) -> Operation:
    """Get the latest operation

    Args:
        tamr: A Tamr client

    Returns:
        The latest job
    """
    op = get_all(tamr)[0]
    return op


def get_details(*, operation: Operation) -> str:
    """Return a text describing the information of a job

    Args:
        operation: A Tamr operation

    Returns:
        A text describing the information of a job
    """
    return (
        f"Host: {operation.client.host} \n Job: {operation.resource_id} \n "
        f"Description: {operation.description} \n Status: {operation.state} "
    )


def get_all(tamr: Client) -> List[Operation]:
    """Get a list of all jobs or operations.

    Args:
        tamr: A Tamr client

    Returns:
        A list of Operation objects.

    """
    response = tamr.get(
        "/api/versioned/v1/operations", headers={"Accept": "application/json"}, stream=True
    ).json()
    ops = [Operation.from_json(tamr, item) for item in response]

    return ops


def get_active(tamr: Client) -> List[Operation]:
    """Get a list of pending and running jobs.

    Args:
        tamr: A Tamr client

    Returns:
        A list of Operations objects

    """
    ops = get_all(tamr)
    active_states = [OperationState.PENDING, OperationState.RUNNING]
    active_ops = [op for op in ops if OperationState(op.state) in active_states]

    return active_ops


def wait(
    operation: Operation, *, poll_interval_seconds: int = 3, timeout_seconds: Optional[int] = None
) -> Operation:
    """Continuously polls for this operation's server-side state.

    Args:
        operation: Operation to be polled.
        poll_interval_seconds: Time interval (in seconds) between subsequent polls.
        timeout_seconds: Time (in seconds) to wait for operation to resolve.

    Raises:
        TimeoutError: If operation takes longer than `timeout_seconds` to resolve.
    """
    started = now()
    while timeout_seconds is None or now() - started < timeout_seconds:
        if operation.status is None:
            return operation
        elif operation.status["state"] in [OperationState.PENDING, OperationState.RUNNING]:
            sleep(poll_interval_seconds)
        elif operation.status["state"] in [
            OperationState.CANCELED,
            OperationState.SUCCEEDED,
            OperationState.FAILED,
        ]:
            return operation
        operation = operation.poll()
    raise TimeoutError(f"Waiting for operation took longer than {timeout_seconds} seconds.")


def monitor(
    operation: Operation, *, poll_interval_seconds: float = 1, timeout_seconds: float = 300
) -> Operation:
    """Continuously polls for this operation's server-side state and returns operation
    when there is a state change

    Args:
        operation: Operation to be monitored.
        poll_interval_seconds: Time interval (in seconds) between subsequent polls.
        timeout_seconds: Time (in seconds) to wait for operation to resolve.

    Raises:
        TimeoutError: If operation takes longer than `timeout_seconds` to resolve.
    """
    status = OperationState[operation.state]
    started = now()
    while timeout_seconds is None or now() - started < timeout_seconds:
        operation = operation.poll()
        new_status = OperationState[operation.state]
        if operation.status is None:
            return operation
        elif new_status == status:
            sleep(poll_interval_seconds)
        else:
            return operation
    raise TimeoutError(f"Waiting for operation took longer than {timeout_seconds} seconds.")


def safe_estimate_counts(project) -> Operation:
    """
    Run the estimate counts job of project that works if it is the first
    Args:
        project: A Tamr project object
    Returns:
        An operation object for the estimate pairs job
    """
    project = project.as_mastering()
    response = project.client.post(f"{project.api_path}/estimatedPairCounts:refresh")

    return Operation.from_response(project.client, response).wait()



"""Enum representing job status"""
from enum import IntEnum
import logging
from tamr_unify_client.operation import Operation

import tamr_toolbox

LOGGER = logging.getLogger(__name__)


class PlanNodeStatus(IntEnum):
    """
    A class representing job status for executing a concurrent pipeline
    """

    FAILED = -2  # one or more of the jobs for the project failed
    CANCELED = -1  # one or more of the jobs for the project was cancelled
    BLOCKED = 0  # running this project is blocked by upstream dependencies failing
    PLANNED = 1  # default state - planned but not yet runnable due to upstream dependencies
    SKIPPABLE = 2  # used when, for example, starting later in the plan
    RUNNABLE = 3  # all dependencies met (e.g. starting tier project) - the project can be run
    PENDING_NEXT_STEP = 4  # project has run some steps successfully and has more steps remaining
    RUNNING = 5  # a job for the project is currently running
    SUCCEEDED = 6  # all project steps completed successfully


def from_tamr_op(op: Operation) -> PlanNodeStatus:
    LOGGER.debug(f"checking state of operation {op}")
    # check each operation
    if op.state == "PENDING" or op.state == "RUNNING":
        return PlanNodeStatus.RUNNING
    elif op.state == "SUCCEEDED":
        return PlanNodeStatus.SUCCEEDED
    elif op.state == "FAILED":
        return PlanNodeStatus.FAILED
    elif op.state == "CANCELED" or op.state == "CANCELING":
        return PlanNodeStatus.CANCELED
    else:
        raise ValueError(f"{op.state} can't be translated to a PlanNodeStatus object")


def from_plan_node(
    plan_node: "tamr_toolbox.workflow.concurrent.PlanNode.PlanNode",
) -> PlanNodeStatus:
    """
    Return a PlanNode Status from a PlanNode object.

    Args:
        plan_node: the plan node object to get the status of

    Returns:
        Latest view of the status of the PlanNode passed in
    """
    # get the status of all operations for this node and convert to status
    # if there are none then and there is nothing to check - just return the status
    if plan_node.operations is None:
        return plan_node.status

    # if there are operations check the operations poll them and see the latest state
    LOGGER.debug(f"getting plan node status at plan node at step {plan_node.current_step}")
    operations_to_poll = [x for x in plan_node.operations]
    all_statuses = [from_tamr_op(x.poll()) for x in operations_to_poll]

    # since the enum is ordered just send the minimum status of all operations back
    # this means any failures/cancellations result in failure/cancelled
    # and then any operations that are less than succeeded represent the overall
    # status of this node, which is correct (remember that each node is a project with
    # multiple operations associated with updating it)

    # if the plan node has next steps and its min is succeeded then mark as pending
    min_status = min(all_statuses)
    if plan_node.steps_to_run and min_status == PlanNodeStatus.SUCCEEDED:
        return PlanNodeStatus.PENDING_NEXT_STEP
    else:
        return min_status



"""Tests for common tasks to the testing framework only"""
from types import ModuleType

from tests._common import get_toolbox_root_dir
from pathlib import Path
import os
import importlib


def test__valid_toolbox_root_dir():
    path = get_toolbox_root_dir()
    assert path.exists()
    assert path.is_absolute()
    # test that we can find this file using the toolbox root directory
    assert path / "tests" / "test__common.py" == Path(__file__)


def test__import_namespaces():
    def check_subpackage_imports(subpackage: ModuleType, directory_path: Path) -> None:
        """Recursively asserts that all files/directories within a directory path are importable
        from the subpackage

        Args:
            subpackage: The subpackage to check import from
            directory_path: The directory to compare with the subpackage imports
        """
        print(f"Checking all files/directories within {directory_path} are importable")
        # Collect names of packages we expect based on filenames in the directory
        modules_by_files = {
            f.replace(".py", "") for f in os.listdir(directory_path) if not f.startswith("_")
        }
        # Collect names of packages that are importable from the module
        modules_by_namespace_all = set(subpackage.__all__)
        modules_by_namespace_dir = {p for p in subpackage.__dir__() if (not p.startswith("_"))}

        # We use subset here to allow for imports that are not in the direct file system folder
        # such as the project._common files which are imported through specific project types
        assert modules_by_files.issubset(modules_by_namespace_all)
        assert modules_by_files.issubset(modules_by_namespace_dir)

        # For any subpackages that are also directories, perform the same check
        for sub_module_name in modules_by_namespace_dir:
            sub_module_path = directory_path / sub_module_name
            if os.path.isdir(sub_module_path):
                sub_module = importlib.import_module(f"{subpackage.__name__}.{sub_module_name}")
                check_subpackage_imports(sub_module, sub_module_path)

    import tamr_toolbox

    toolbox_dir = get_toolbox_root_dir() / "tamr_toolbox"

    check_subpackage_imports(tamr_toolbox, toolbox_dir)



"""Tasks related to creation of Email notifications"""
import logging
import smtplib
import ssl

from typing import Union, List, Optional, Tuple
from tamr_toolbox.models.data_type import JsonDict
from email.mime.text import MIMEText
from tamr_toolbox.notifications.common import _monitor_job as monitor_job_common

from tamr_unify_client import Client
from tamr_unify_client.operation import Operation

from tamr_toolbox.models.operation_state import OperationState
from tamr_toolbox.utils.operation import get_details

LOGGER = logging.getLogger(__name__)


def _build_message(*, message: str, subject_line: str, sender: str, recipients: List[str]) -> str:
    """Builds email message in Multipurpose Internet Mail Extensions (MIME) format. MIME is an
    Internet standard that extends the format of email messages

    Args:
        message: Body of email message
        subject_line: subject of email
        sender: email address of sender
        recipients: list of emails to send message to

    Returns:
        Email as a string
    """

    # build email
    msg = MIMEText(message)
    msg["Subject"] = subject_line
    msg["From"] = sender
    msg["To"] = ", ".join(recipients)
    return msg.as_string()


def send_email(
    *,
    message: str,
    subject_line: str,
    sender_address: str,
    sender_password: str,
    recipient_addresses: List[str],
    smtp_server: str,
    smtp_port: str,
    use_tls: bool = True,
    keyfile: Optional[str] = None,
    certfile: Optional[str] = None,
) -> JsonDict:
    """Sends a message via email to list of recipients

    Args:
        message: Body of email message
        subject_line: subject of email
        sender_address: email address to send message from ex: my_pipeline@gmail.com
        sender_password: password to login to sender_email
        recipient_addresses: list of emails to send message to ex: [client_email@gmail.com]
        smtp_server: smtp server address of sender_email ex: smtp.gmail.com
        smtp_port: port to send email from, use 465 for SSL, use 587 for TLS
        use_tls: A boolean value to turn on/off TLS protocol
        keyfile: the private key to a TLS/SSL certificate, usually PEM format
        certfile: TLS/SSL cert file issued by a Certificate Authority (CA), usually PEM format

    Returns:
        A dict with the message and response codes from the smtp server if there are any
        errors. The dictionary will contain one entry for each recipient that was refused.
        Each entry contains a tuple of the SMTP error code and the accompanying error
        message sent by the server. A successful response will contain an empty dict.

    Raises:
        SMTPException. The base exception class used by the smtplib module
    """
    # build email
    msg = _build_message(
        message=message,
        subject_line=subject_line,
        sender=sender_address,
        recipients=recipient_addresses,
    )
    response = dict()
    response["message"] = message

    context = ssl.create_default_context()
    with smtplib.SMTP(smtp_server, smtp_port) if use_tls else smtplib.SMTP_SSL(
        smtp_server, smtp_port, keyfile=keyfile, certfile=certfile, context=context
    ) as server:
        if use_tls:
            server.starttls(keyfile=keyfile, certfile=certfile, context=context)

        # login and send message
        server.login(sender_address, sender_password)
        errors = server.sendmail(sender_address, recipient_addresses, msg)
        response["errors"] = errors

    return response


def _send_job_status_message(
    *,
    sender_address: str,
    sender_password: str,
    recipient_addresses: List[str],
    smtp_server: str,
    smtp_port: str,
    operation: Operation,
    notify_states: List[OperationState],
    use_tls: bool = False,
    keyfile: Optional[str] = None,
    certfile: Optional[str] = None,
) -> JsonDict:
    """Checks operation state and if in `notify_states` sends the message.

    Args:
        sender_address: email address to send message from ex: my_pipeline@gmail.com
        sender_password: password to login to sender_email
        recipient_addresses: list of emails to send message to ex: [client_email@gmail.com]
        smtp_server: smtp server address of sender_email ex: smtp.gmail.com
        smtp_port: port to send email from, use 465 for SSL
        operation: A Tamr Operation
        notify_states: States for which notifications should be sent
        use_tls: A boolean value to opt to use TLS protocol
        keyfile: the private key to a TLS/SSL certificate, usually PEM format
        certfile: TLS/SSL cert file issued by a Certificate Authority (CA), usually PEM format

    Returns:
        A dict with the message and response codes from the smtp server if there are any
        errors. The dictionary will contain one entry for each recipient that was refused.
        Each entry contains a tuple of the SMTP error code and the accompanying error
        message sent by the server. A successful response will contain an empty dict.
    """
    state = OperationState[operation.state]
    message, resp = None, None
    if state in notify_states:
        message = get_details(operation=operation)
        resp = send_email(
            message=message,
            subject_line=f"Job {operation.resource_id}: {state}",
            sender_address=sender_address,
            sender_password=sender_password,
            recipient_addresses=recipient_addresses,
            smtp_server=smtp_server,
            smtp_port=smtp_port,
            use_tls=use_tls,
            keyfile=keyfile,
            certfile=certfile,
        )
    return resp


def monitor_job(
    tamr: Client,
    *,
    sender_address: str,
    sender_password: str,
    recipient_addresses: List[str],
    smtp_server: str,
    smtp_port: str,
    operation: Union[int, str, Operation],
    poll_interval_seconds: float = 1,
    timeout_seconds: Optional[float] = None,
    notify_states: Optional[List[OperationState]] = None,
    use_tls: bool = False,
    keyfile: Optional[str] = None,
    certfile: Optional[str] = None,
) -> List[Tuple[str, JsonDict]]:
    """Monitors a Tamr Operation and sends an email when the job status is updated

    Args:
        tamr: A Tamr client
        sender_address: email address to send message from ex: my_pipeline@gmail.com
        sender_password: password to login to sender_email
        recipient_addresses: list of emails to send message to ex: [client_email@gmail.com]
        smtp_server: smtp server address of sender_email ex: smtp.gmail.com
        smtp_port: port to send email from, use 465 for SSL
        operation: A job ID or a Tamr operation
        poll_interval_seconds: Time interval (in seconds) between subsequent polls
        timeout_seconds: Time (in seconds) to wait
        notify_states : States for which notifications should be sent, use None for all states
        use_tls: A boolean value to opt to use TLS protocol
        keyfile: the private key to a TLS/SSL certificate, usually PEM format
        certfile: TLS/SSL cert file issued by a Certificate Authority (CA), usually PEM format

    Returns:
        A list of dicts. Each dict comtains the message and error response codes from the smtp
        server. The dictionary will contain one entry for each recipient that was refused.
        Each entry contains a tuple of the SMTP error code and the accompanying error
        message sent by the server. A successful response will contain an empty dict.
    """
    list_responses = monitor_job_common(
        tamr=tamr,
        send_message=send_email,
        send_status_function=_send_job_status_message,
        sender_address=sender_address,
        sender_password=sender_password,
        recipient_addresses=recipient_addresses,
        smtp_server=smtp_server,
        smtp_port=smtp_port,
        operation=operation,
        poll_interval_seconds=poll_interval_seconds,
        timeout_seconds=timeout_seconds,
        notify_states=notify_states,
        use_tls=use_tls,
        keyfile=keyfile,
        certfile=certfile,
    )

    return list_responses



"""Tasks related to handling jdbc information for the Tamr auxiliary service DF-connect"""
from dataclasses import dataclass
from tamr_toolbox.models.data_type import JsonDict


@dataclass
class JdbcInfo:
    """
    A dataclass to tie together relevant data to ingest data into df_connect.

    Args:
        jdbc_url: The jdbc url for the database to which to connect
        db_user: The database user as whom to authenticate
        db_password: The password for the database user
        fetch_size: The number of records by which to chunk the jdbc ResultSet
    """

    jdbc_url: str
    db_user: str
    db_password: str
    fetch_size: int


def from_config(
    config: JsonDict, *, config_key: str = "df_connect", jdbc_key: str = "ingest"
) -> JdbcInfo:
    """Create an instance of JdbcInfo from a json object.

    Args:
        config: A json dictionary containing configuration values
        config_key: the top-level key of the config to use.
        jdbc_key: the key to use for the jdbc block. Needs to be within `config_key` block.
                  Defaults to 'ingest', but can be used to specify any sub-block of a config
                  object or yaml file. See example configs and exports for more context.
    """
    return JdbcInfo(
        jdbc_url=config[config_key]["jdbc"][jdbc_key]["jdbc_url"],
        db_user=config[config_key]["jdbc"][jdbc_key]["db_user"],
        db_password=config[config_key]["jdbc"][jdbc_key]["db_password"],
        fetch_size=config[config_key]["jdbc"][jdbc_key]["fetch_size"],
    )


def create(*, jdbc_url: str, db_user: str, db_password: str, fetch_size: int) -> JdbcInfo:
    """
    A simple wrapper to create an object of type `JdbcInfo`

    Args:
        jdbc_url: The jdbc url for the database to which to connect
        db_user: The database user as whom to authenticate
        db_password: The password for the database user
        fetch_size: The number of records by which to chunk the jdbc ResultSet

    Returns:
        An instance of a JdbcInfo object.
    """
    return JdbcInfo(
        jdbc_url=jdbc_url, db_user=db_user, db_password=db_password, fetch_size=fetch_size
    )



"""Functions related to projects upstream of a specified project"""
from typing import List, Optional
import logging

from tamr_unify_client.project.resource import Project
from tamr_unify_client.dataset.resource import Dataset

LOGGER = logging.getLogger(__name__)


def _find_projects(
    project: Project,
    all_projects: Optional[List[Project]] = None,
    project_list: Optional[List[str]] = None,
    upstream_projects: Optional[List[str]] = None,
) -> List[Project]:
    """Check for upstream projects that have source datasets that require
    updating
    Args:
        project: a tamr project
        track of projects that need to be checked for upstream dependencies
        all_projects: list of all projects that exist within Tamr
        project_list: keeps track of projects that have yet to be checked
        for upstream dependencies.
        upstream_projects: list to keep track of projects within a chain of projects
    Returns:
        List of tamr projects in a chained workflow
    """
    # Obtain the name of project to be updated and initiate the list of
    # projects (project_list) that are maintained within the check_for_upstream_project
    # function to check for upstream dependencies

    # Get project_id from project
    project_id = project.resource_id

    if not upstream_projects:
        upstream_projects = []
        project_list = [project_id]

        all_projects = {}

        for project in project.client.projects:
            all_projects[project.name] = project.resource_id

    for project_id in project_list:
        # Obtain resource id of project to be updated from the dictionary of
        # project details
        project = project.client.projects.by_resource_id(project_id)
        if project.resource_id not in upstream_projects:
            upstream_projects.append(project.resource_id)

        # Remove the current project from the list tracking projects that
        # need to be reviewed for upstream projects/dependencies
        project_list.remove(project_id)

        project_source_datasets = project.client.get(
            f"/api/versioned/v1/projects/{project.resource_id}/inputDatasets"
        ).json()
        relative_ids = [dataset["relativeId"] for dataset in project_source_datasets]

        # Check which upstream projects the upstream datasets derive from
        for relative_id in relative_ids:
            dataset_details = project.client.datasets.by_relative_id(relative_id)

            # If these projects are linked to further datasets upstream,
            # keep track of these upstream projects in project_list
            upstream_datasets = dataset_details.usage().usage.output_from_project_steps
            if upstream_datasets:
                for upstream_dataset in upstream_datasets:
                    upstream_project = upstream_dataset.project_name
                    project_id = all_projects[upstream_project]
                    if project_id not in project_list:
                        project_list.append(project_id)

    # If there are projects added to the project list, check whether these
    # projects have dependencies upstream by rerunning the
    # check_for_upstream_datasets recursively
    if project_list:
        return _find_projects(
            project=project,
            project_list=project_list,
            all_projects=all_projects,
            upstream_projects=upstream_projects,
        )
    else:
        # If the first project in the chain is reached, return the list of
        # projects that need to be updated

        # Reverse list of projects so that the first project in the chain is
        # returned first
        upstream_projects.reverse()

        upstream_projects = [
            project.client.projects.by_resource_id(project_id) for project_id in upstream_projects
        ]

        return upstream_projects


def _find_datasets(dataset: Dataset) -> List[Dataset]:
    """Returns a dataset's upstream datasets.

    Args:
        dataset: a Tamr Dataset Object
    Returns:
        List of upstream datasets

    """

    # Create empty list to populate with upstream datasets
    data_upstream = []
    # Main dataset
    main_dataset = dataset
    # Create list of datasets to go through
    datasets_to_check = [main_dataset]
    # Create list of checked datasets
    datasets_checked = []

    # Find all upstream datasets
    while len(datasets_to_check) > 0:
        # output is a list; check if anything present, add output to upstream datasets
        # and add to list of datasets to check
        if datasets_to_check[0].name not in datasets_checked:
            upstream = _request_upstream_datasets(datasets_to_check[0])
            datasets_checked.append(datasets_to_check[0].name)
            datasets_to_check.remove(datasets_to_check[0])
            if len(upstream) != 0:
                data_upstream.extend(upstream)
                datasets_to_check.extend(upstream)
        else:
            datasets_to_check.remove(datasets_to_check[0])
    # Deduplicate & collect names: deduplicate based on identical names
    dataset_names = []
    for data in data_upstream:
        dataset_names.append(data.name)
    # Use dictionary to remove duplicate names
    names_datasets_zip = dict(zip(dataset_names, data_upstream))
    dataset_upstream = list(names_datasets_zip.values())
    return dataset_upstream


def _request_upstream_datasets(dataset: Dataset) -> Dataset:

    """Returns a dataset's upstream dataset

    Args:
        dataset: a Tamr Dataset Object
    Returns:
        The upstream datasets
    """
    # Find upstream datasets, output is a DatasetURI
    upstream = dataset.upstream_datasets()
    dataset_upstream = []
    # Make Dataset our of DatasetURI
    for data in upstream:
        dataset_upstream.append(dataset.client.datasets.by_resource_id(data.resource_id))
    return dataset_upstream


def datasets(dataset: Dataset) -> List[Dataset]:
    """Check for upstream datasets associated with a specified dataset

    Args:
        dataset: the Tamr dataset for which associated upstream datasets are retrieved
    Returns:
            List of Tamr datasets upstream of the target dataset
    """
    upstream_datasets = _find_datasets(dataset)

    return upstream_datasets


def projects(project: Project) -> List[Project]:
    """Check for upstream projects associated with a specified project

    Args:
        project: the tamr project for which associated upstream projects are retrieved
    Returns:
            List of tamr projects upstream of the target project
    """

    # Retrieve upstream projects (if any) from a specified Project
    upstream_projects = _find_projects(project)

    return upstream_projects



"""Tasks related to testing code"""
import logging
import json
import os
import re
from inspect import getfile

from pathlib import Path
from time import sleep
from typing import Callable, List, Optional, Union, Dict
from unittest import mock
from functools import wraps
from requests import Response, ConnectionError

from tamr_unify_client.operation import Operation

from tamr_toolbox import utils


LOGGER = logging.getLogger(__name__)

WINDOWS_RESERVED_CHARACTER_MAP = {
    "<": "lt",
    ">": "gt",
    ":": "colon",
    '"': "dquote",
    "/": "fslash",
    "\\": "bslash",
    "|": "pipe",
    "?": "qmark",
    "*": "asterisk",
}


def _response_to_json(resp: Response, ip_dict: Dict[str, int]) -> str:
    """Converts a Response object into json string readable by the responses mocking library
    Args:
        resp: Response from a Tamr API call
        ip_dict: Mapping of previously encountered IP addresses to their anonymization number
    Returns:
        The response represented as a json string
    """
    if resp.encoding is None:
        resp.encoding = "utf-8"
    resp_log = {
        "method": resp.request.method,
        "url": _anonymize_url(resp.request.url, ip_dict),
        "status": resp.status_code,
        "content_type": resp.headers.get("Content-Type"),
        "body": resp.text,
    }
    return json.dumps(resp_log, ensure_ascii=False)


def _anonymize_url(url: str, ip_dict: Dict[str, int]) -> str:
    """Returns a anonymized url. Updates the dictionary inplace if a new ip is encountered

    Args:
        url: A URL
        ip_dict: Previously encountered IP addresses and an assigned numeric value

    Returns:
        URL with the IP address anonymized
    """
    regex_match = re.match(r"(?i)(^https?://)(.*?)([/:].*$)", url)
    ip = regex_match.group(2)

    try:
        num = ip_dict[ip]
    except KeyError:
        ip_dict[ip] = len(ip_dict.values()) + 1
        num = ip_dict[ip]

    return f"{regex_match.group(1)}ip-{num:05d}{regex_match.group(3)}"


def _collect_operation_calls(
    *, response: Response, poll_interval_seconds: int = 3
) -> List[Response]:
    """If the provided response is an Operation, wait for the operation to complete and
    return responses related to that operation for cases where we are running synchronously.

    Args:
        response: A previous Response generated from the same Tamr client
        poll_interval_seconds: Time interval (in seconds) between subsequent polls

    Returns:
        Responses related to polling the operation

    """
    client = utils.client._from_response(response)
    op = Operation.from_response(client, response)
    LOGGER.info(f"Waiting for operation to complete: {op}")

    request_while_pending = client.get(endpoint=f"/api/versioned/v1/operations/{op.resource_id}")

    while op.state == "PENDING":
        op = op.poll()
        sleep(poll_interval_seconds)
    request_while_running = client.get(endpoint=f"/api/versioned/v1/operations/{op.resource_id}")

    op.wait()
    request_when_complete = client.get(endpoint=f"/api/versioned/v1/operations/{op.resource_id}")

    return [request_while_pending, request_while_running, request_when_complete]


def _log_response(
    *, log_path: Path, ip_dict: Dict[str, int], response: Response, asynchronous: bool = False
) -> None:
    """Appends a response to a file. If the response returned is
     a Tamr Operation, poll the operation until complete and log those responses as well
     unless asynchronous is set to True.

    Args:
        log_path: File to write the response to
        ip_dict: Mapping of previously encountered IP addresses to their anonymization number
        response: The response to log
        asynchoronous: Whether or not to `wait` for Operations called during the running of tests

    """
    LOGGER.info(f"logged request: {response.url}")
    with log_path.open(mode="a", encoding="utf-8") as f:
        all_responses = [response]

        # Poll and wait for operations, if applicable
        is_operation_request = bool(
            re.match(re.compile(".*/api/versioned/v1/operations/.*"), response.url)
        )
        is_get_request = response.request.method == "GET"
        is_not_error = response.ok

        if is_get_request and is_operation_request and is_not_error and not asynchronous:
            wait_resp = _collect_operation_calls(response=response)
            all_responses.extend(wait_resp)

        all_json = [_response_to_json(r, ip_dict) for r in all_responses]
        f.writelines([f"{j}\n" for j in all_json])


def _build_response_log_path(
    *, test_func: Callable, response_logs_dir: Optional[Union[str, Path]], **kwargs
) -> Path:
    """Returns a file path for API response logs for a given test and test parameters

    Args:
        test_func: The test function
        **kwargs: Arguments to the test function

    Returns:
        File path for the API response logs

    """
    # Convert test arguments and their values to a string, skipping ignored arguments
    test_params = "_".join([f"{k}={v}" for k, v in {**kwargs}.items()])
    # Remove reserved characters from ndjson name
    for char in WINDOWS_RESERVED_CHARACTER_MAP:
        test_params = test_params.replace(char, WINDOWS_RESERVED_CHARACTER_MAP[char])

    if len(test_params) > 0:
        test_params = "__" + test_params

    if response_logs_dir is None:
        # If no directory is provided, create a directory with the name of the test file
        # in a directory called "response_logs" located in the same directory as the test file
        dir_matcher = re.match(r"(.*)(?:\\|/)(.*).py", str(Path(getfile(test_func))))

        response_logs_dir = f"{dir_matcher.group(1)}/response_logs/{dir_matcher.group(2)}"

    return Path(f"{response_logs_dir}/{test_func.__name__}{test_params}.ndjson")


def mock_api(
    *,
    response_logs_dir: Optional[Union[str, Path]] = None,
    enforce_online_test: bool = False,
    asynchronous: bool = False,
) -> Callable:
    """Decorator for `pytest` tests that mocks API requests by reading a file of
    pre-generated responses. Will generate responses file based on a real connection
    if pre-generated responses are not found.

    Args:
        response_logs_dir: Directory to read/write response logs
        enforce_online_test: Whether an online test should be run, even if a response log
            already exists
        asynchronous: Whether or not to `wait` for Operations called during the running of tests

    Returns:
        Decorated function

    """

    def wrap(test_function: Callable):
        @wraps(test_function)
        def wrapped(**kwargs):

            response_log_path = _build_response_log_path(
                test_func=test_function, response_logs_dir=response_logs_dir, **kwargs
            )

            if response_log_path.exists() and enforce_online_test:
                # Delete the file to enforce an online test
                response_log_path.unlink()

            if response_log_path.exists():
                try:
                    LOGGER.info(f"Running offline test based on file at {response_log_path}")
                    _run_offline_test(
                        response_log_path=response_log_path, test_function=test_function, **kwargs
                    )
                except ConnectionError as e:
                    msg = (
                        f"A required API call was missing from response logs file for this "
                        f"offline test ({response_log_path}). The response log file must be "
                        f"regenerated. Delete the existing file to automatically regenerate a "
                        f"new one. Caused by: {e}"
                    )
                    LOGGER.error(msg)
                    raise ConnectionError(e)
            else:
                _run_online_test(
                    response_log_path=response_log_path,
                    test_function=test_function,
                    asynchronous=asynchronous,
                    **kwargs,
                )

        return wrapped

    return wrap


# Handle ModuleNotFoundError to allow tamr_toolbox to be used when the optional dependency
# `responses` is not installed
try:
    import responses

    # Stores the original _real_send function of requests
    _BASE_FIND_MATCH = responses.RequestsMock._find_match
    # Stores the original _real_send function of responses
    _BASE_SEND_REAL = responses._real_send

    @responses.activate
    @mock.patch.object(Operation.wait, "__defaults__", (0, None))  # sets operation wait time to 0
    def _run_offline_test(response_log_path: Path, test_function: Callable, **kwargs) -> None:
        """Runs a test function against saved API responses located in a file

        Args:
            response_log_path: Location of saved API responses
            test_function: The function to test
            **kwargs: Keyword arguments for the test function

        """
        with response_log_path.open(encoding="utf-8") as f:
            for line in f:
                response = json.loads(line)
                responses.add(**response)

        ip_lookup = {}

        def _find_anonymized_match(self, request):
            """Allows responses library to match requests for an ip address to match to an
            anonymized ip address
            """
            request.url = _anonymize_url(request.url, ip_lookup)
            return _BASE_FIND_MATCH(self, request)

        with mock.patch("responses.RequestsMock._find_match", new=_find_anonymized_match):
            test_function(**kwargs)

    @responses.activate
    def _run_online_test(
        response_log_path: Path, test_function: Callable, asynchronous: bool = False, **kwargs
    ) -> None:
        """Runs a test function against a Tamr instance and saves the API responses to a file

        Args:
            response_log_path: Location to save API responses
            test_function: The function to test
            asynchoronous: Whether or not to run asynchronously
                (i.e. don't `wait` for operations to finish)
            **kwargs: Keyword arguments for the test function

        """
        LOGGER.info(
            f"Online test running against Tamr instance. "
            f"Creating new file at {response_log_path}. This may take a while ..."
        )
        os.makedirs(response_log_path.parent, exist_ok=True)
        response_log_path.touch()

        # Each time an API call is made, allow it to pass through responses and make a real call
        # Each time a real call is made, log the response in the response file
        responses.add_passthru(re.compile(".*"))

        ip_lookup = {}

        def _send_real_with_log(*args, **kwargs) -> Response:
            """Logs the response from BASE_SEND_REAL

            Args:
                *args: The positional arguments for BASE_SEND_REAL
                **kwargs: The keyword arguments for BASE_SEND_REAL

            Returns:
                The response from the call

            """
            response = _BASE_SEND_REAL(*args, **kwargs)

            # Prevent recursion
            with mock.patch("responses._real_send", new=_BASE_SEND_REAL):
                _log_response(
                    log_path=response_log_path,
                    response=response,
                    ip_dict=ip_lookup,
                    asynchronous=asynchronous,
                )
            return response

        with mock.patch("responses._real_send", new=_send_real_with_log):
            test_function(**kwargs)

        # Setting the passthru above permanently changes state for online testing
        # Reset passthru to default
        responses.mock.passthru_prefixes = ()
        responses._default_mock.passthru_prefixes = ()

except ModuleNotFoundError as err:
    # Ensure exception is due to responses package being missing
    if err.msg != "No module named 'responses'":
        raise err

    def _run_offline_test(*args, **kwargs):
        """Dummy function to raise the appropriate exception if the function is called without the
        necessary package installed
        """
        import responses  # noqa: F401

    def _run_online_test(*args, **kwargs):
        """Dummy function to raise the appropriate exception if the function is called without the
        necessary package installed
        """
        import responses  # noqa: F401
